using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.CompilerServices;

namespace CS_asynplex
{
	class Message {
		public string sender { get; private set; }
		public string receiver { get; private set; }
		public string tag { get; private set; }
		public object payload { get; private set; }

		public Message(string sender, string receiver, string tag, object payload) {
			this.sender = sender;
			this.receiver = receiver;
			this.tag = tag;
			this.payload = payload;
		}
	}
	


    class Communication
    {
		static List<Message> messages = new List<Message>();
		static bool printSends = true;
		static bool printReceives = true;
		static object messageLock = new object();
		public static readonly object printLock = new object();

		public static void send(Message message) {
			lock (messageLock) {
				messages.Add(message);
			}
			if (printSends) {
				lock (printLock) {
					Console.ForegroundColor = ConsoleColor.DarkGray;
					Console.WriteLine("COMM: {0} sends to {1} (tag: '{2}', msg type: {3}) (queue: {4})", message.sender, message.receiver, message.tag, message.payload == null ? "(null)" : message.payload.GetType().ToString(), messages.Count);
					Console.ForegroundColor = ConsoleColor.Gray;
				}
			}
		}

		public static Message receive(string receiver) {
			return receive(receiver, null, null);
		}

		public static Message receive(string receiver, string sender) {
			return receive(receiver, sender, null);
		}
		
		public static Message receive(string receiver, string sender, string tag) {
			lock (messageLock) {
				for (int i = 0; i < messages.Count; ++i) {
					if (messages[i].receiver == receiver && (sender == null || messages[i].sender == sender || (sender == "I*" && messages[i].sender[0] == 'I')) && (tag == null || messages[i].tag == tag)) {
						Message message = messages[i];
						messages.RemoveAt(i);
						if (printReceives) {
							lock (printLock) {
								Console.ForegroundColor = ConsoleColor.DarkGray;
								Console.WriteLine("COMM: {0} receives from {1} (tag: '{2}', msg type: {3}) (queue: {4})", message.receiver, message.sender, message.tag, message.payload == null ? "(null)" : message.payload.GetType().ToString(), messages.Count);
								Console.ForegroundColor = ConsoleColor.Gray;
							}
						}
						return message;
					}
				}
			}
			return null;
		}
		
		public static void printMessages() {
			lock (messageLock) {
				lock (printLock) {
					Console.Write("Remaining messages (some messages may subsequently be read before all threads terminate):");
					foreach (Message msg in messages) {
						Console.Write(" " + msg.tag);
					}
					Console.WriteLine();
				}
			}
		}
	}
}
