using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Threading;

namespace CS_asynplex {
	class InverseMatrixMessage {
		public int basisNumber;
		public Matrix inverse;
	}

	class InversionProcess : AsynplexProcess {
		private Matrix B, N;
		int numIterationProcs;
		int lastReinversionBasis = 0;
		int basisNumber = 0;

		public InversionProcess(Matrix N, int numIterationProcs)
			: base("V") {
			this.N = new Matrix(N);
			this.B = new Matrix(N.Rows, N.Rows, true, "B");
			this.numIterationProcs = numIterationProcs;
		}

		public override void run() {
			Message message;

			while (!Program.Finished) {
				// V1 <- I8
				while ((message = receive()) != null || basisNumber == lastReinversionBasis) {
					if (Program.Finished)
						return;
					if (message == null) {
						Thread.Sleep(10);
						continue;
					}
					Debug.Assert(message.tag == "I8->V1");
					int[] swap = (int[])message.payload;
					int entering = swap[0];
					int leaving = swap[1];
					Matrix tempCol = B.getColumn(leaving);
					B.setColumn(leaving, N.getColumn(entering));
					N.setColumn(entering, tempCol);
					++basisNumber;
					print(message.sender + " says to let " + leaving + " leave and " + entering + " enter; now at basis " + basisNumber);
				}
				if (Program.Finished)
					return;

				print("Calculating inverse for basis " + basisNumber);
				Matrix BInverse = B.invert();
				lastReinversionBasis = basisNumber;
				//Thread.Sleep(8000);//TODO:remove
				print("Done with inverse for basis " + basisNumber);
				
				// V2 -> I1
				for (int i = 0; i < numIterationProcs; ++i) {
					//TODO: Must also tell what basis this is for
					InverseMatrixMessage invMsg = new InverseMatrixMessage();
					invMsg.basisNumber = basisNumber;
					invMsg.inverse = BInverse;
					send("I" + i, "V2->I1", invMsg);
				}
			}
		}
	}
}
