using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Globalization;
using System.Threading;

namespace CS_asynplex {
	class PrintControl {
		public bool B = false;
		public bool N = false;
		public bool cBasic = false;
		public bool cNonbasic = false;
		public bool b = false;
		public bool piTranspose = false;
		public bool reducedCosts = false;
		public bool stepDirection = false;
		public bool stepLength = false;
		public bool pivotDecisions = false;
		public bool pivotResults = false;
		public bool variableLists = false;
		public bool iterationCount = false;
		public bool iterationSeparator = false;
		public bool objectiveValue = false;
		public bool etaFileSize = false;
		public bool inverseDiscrepancy = false;
	}

	class Program {
		const float EPS = 0.0000005f;

		private static bool equalsEpsilon(float a, float b) {
			return Math.Abs(a - b) < EPS;
		}

		private static object finishedLock = new object();
		private static bool _finished = false;
		public static bool Finished {
			get {
				lock (finishedLock) {
					return _finished;
				}
			}
			set {
				lock (finishedLock) {
					_finished = value;
				}
			}
		}

		static void Main(string[] args) {
			StreamReader sr = new StreamReader(args[0]);
			string[] items = sr.ReadLine().Split(' ');
			//Random random = new Random();

			int m = int.Parse(items[0]) - 1;
			int n = int.Parse(items[1]) - 1;
			Matrix N = new Matrix(m, n, "N");
			Matrix B = new Matrix(m, m, true, "B");
			List<Matrix> etaFile = new List<Matrix>();
			Matrix cNonbasic = new Matrix(n, 1, "cNonbasic");
			Matrix cBasic = new Matrix(m, 1, "cBasic");
			Matrix basicValues = new Matrix(m, 1, "basicValues"); // Starts out as the b vector; will always equal B^{-1} * b

			items = sr.ReadLine().Split(' ');
			for (int c = 0; c < n; ++c) {
				cNonbasic[c, 0] = float.Parse(items[c], CultureInfo.InvariantCulture.NumberFormat);
			}

			for (int r = 0; r < m; ++r) {
				items = sr.ReadLine().Split(' ');
				for (int c = 0; c < n; ++c) {
					N[r, c] = float.Parse(items[c], CultureInfo.InvariantCulture.NumberFormat);
				}
				basicValues[r, 0] = float.Parse(items[n], CultureInfo.InvariantCulture.NumberFormat);
			}
			sr.Close();

			int[] nonbasicVars = new int[n];
			int[] basicVars = new int[m];
			for (int i = 1; i <= n; ++i)
				nonbasicVars[i - 1] = i;
			for (int i = n + 1; i <= n + m; ++i)
				basicVars[i - n - 1] = i;

			bool useAsynplex = true;
			if (useAsynplex) {
				const int numIterationProcesses = 4;
				InversionProcess invertProcessor = new InversionProcess(N, numIterationProcesses);
				BasisChangeProcess basisChangeManager = new BasisChangeProcess();
				ColumnSelectionProcess columnSelectionManager = new ColumnSelectionProcess(m, n);
				Thread t;
				t = new Thread(invertProcessor.run);
				t.Name = "Invert processor";
				t.Start();
				t = new Thread(basisChangeManager.run);
				t.Name = "Basis change manager";
				t.Start();
				t = new Thread(columnSelectionManager.run);
				t.Name = "Column selection manager";
				t.Start();
				for (int i = 0; i < numIterationProcesses; ++i) {
					IterationProcess iter = new IterationProcess(i, numIterationProcesses, m, n, N, basicValues, cNonbasic);
					t = new Thread(iter.run);
					t.Name = "Iteration process " + i;
					t.Start();
				}
				return;
			}

			Matrix originalBInverse = new Matrix(B.Rows, B.Cols, true, "orig B^I");
			PrintControl printControl = new PrintControl() { iterationCount = true, objectiveValue = true, pivotResults = true};
			int iterationCount = 1;
			int reinversionPeriod = 5;
			while (true) {
				if (printControl.B)			B.print();
				if (printControl.N)			N.print();
				if (printControl.cBasic)	cBasic.print();
				if (printControl.cNonbasic)	cNonbasic.print();
				if (printControl.b)			basicValues.transpose().print();

				Matrix bInverse;
				while (true) {
					try {
						if ((iterationCount - 1) % reinversionPeriod == 0) {
							originalBInverse = B.invert();
							etaFile.Clear();
						}
						if (printControl.etaFileSize) Console.WriteLine("{0} eta matrices", etaFile.Count);
						bInverse = originalBInverse;
						foreach (Matrix eta in etaFile) {
							bInverse = eta * bInverse;
						}
						if (etaFile.Count > 0 && printControl.inverseDiscrepancy) {
							float max = 0;
							Matrix discrepancy = bInverse - B.invert();
							for (int r = 0; r < discrepancy.Rows; ++r) {
								for (int c = 0; c < discrepancy.Cols; ++c) {
									float absVal = Math.Abs(discrepancy[r, c]);
									if (absVal > max)
										max = absVal;
								}
							}
							Console.WriteLine("Largest discrepancy: {0}", max);
						}
						break;
					}
					catch (ArgumentException) {
						throw;
						//Console.WriteLine("reinverting");
						//int i = random.Next(0, m), j;
						//do {
						//    j = random.Next(0, m);
						//} while (j == i);
						//Matrix col = B.getColumn(i);
						//B.setColumn(i, B.getColumn(j));
						//B.setColumn(j, col);
						//continue;
					}
				}

				//BTRAN
				Matrix piTranspose = cBasic.transpose() * bInverse; // \pi^T
				piTranspose.Name = "pi^T";
				if (printControl.piTranspose) piTranspose.print();

				//PRICE
				Matrix reducedCosts = cNonbasic.transpose() - piTranspose * N; // \hat{c}_N^T
				reducedCosts.Name = "reduced costs";
				if (printControl.reducedCosts) reducedCosts.print();

				if (printControl.objectiveValue) {
					Matrix objectiveValue = cBasic.transpose() * basicValues;
					Console.WriteLine("Objective value: {0}", objectiveValue[0, 0]);
				}

				//CHUZC
				int enteringCol = -1; // q
				for (int c = 0; c < reducedCosts.Cols; ++c) {
					if (reducedCosts[0, c] > 0) {
						Console.Write(c + " ");
					}
					if (reducedCosts[0, c] > 0 
						&& (enteringCol == -1 
							|| (reducedCosts[0, c] > reducedCosts[0, enteringCol] && !equalsEpsilon(reducedCosts[0, c], reducedCosts[0, enteringCol]))
							|| (equalsEpsilon(reducedCosts[0, c], reducedCosts[0, enteringCol]) && nonbasicVars[c] < nonbasicVars[enteringCol]))) {
						if (printControl.pivotDecisions) Console.WriteLine("Choosing {0} over {1} to enter; reduced cost is {2}", c + 1, enteringCol + 1, reducedCosts[0, c]);
						enteringCol = c;
					}
				}
				Console.WriteLine();
				if (enteringCol == -1) {
					Console.WriteLine("Optimal");
					break;
				}
				int enteringVar = nonbasicVars[enteringCol];
				if (printControl.pivotResults) Console.WriteLine("Entering variable: x{1} (column {0})", enteringCol, enteringVar);

				//FTRAN
				Matrix stepDirection = bInverse * N.getColumn(enteringCol); // \hat{a}_q
				stepDirection.Name = "step direction";
				if (printControl.stepDirection) stepDirection.print();

				//CHUZR
				int leavingRow = -1; // p
				float largestRatio = 0; // step length (\alpha)
				for (int r = 0; r < basicValues.Rows; ++r) {
					float ratio;
					if (stepDirection[r, 0] == 0 && basicValues[r, 0] == 0)
						ratio = 0;
					else
						ratio = stepDirection[r, 0] / basicValues[r, 0];
					if (ratio > 0 && (leavingRow == -1 || ratio > largestRatio || (ratio == largestRatio && basicVars[r] < basicVars[leavingRow]))) {
						if (printControl.pivotDecisions) Console.WriteLine("Choosing {0} over {1} to leave; ratio is {2}", r + 1, leavingRow + 1, ratio);
						leavingRow = r;
						largestRatio = ratio;
					}
				}
				if (leavingRow == -1) {
					Console.WriteLine("Unbounded");
					break;
				}
				float stepLength = 1 / largestRatio;
				int leavingVar = basicVars[leavingRow];
				if (printControl.pivotResults) Console.WriteLine("Leaving variable: x{1} (row {0})", leavingRow, leavingVar);
				if (printControl.stepLength) Console.WriteLine("Step length: {0}", stepLength);

				//UPRHS
				basicValues -= stepLength * stepDirection;
				basicValues[leavingRow, 0] = stepLength;
				basicValues.Name = "b";
				
				//UPDATE (should normally also calculate/update inverse here)
				Matrix tempM = N.getColumn(enteringCol);
				N.setColumn(enteringCol, B.getColumn(leavingRow));
				B.setColumn(leavingRow, tempM);

				float dxi = stepDirection[leavingRow, 0];
				if (dxi == 0) throw new Exception();
				Matrix etaMatrix = new Matrix(B.Rows, B.Cols, true, "Eta" + etaFile.Count);
				for (int r = 0; r < etaMatrix.Rows; ++r) {
					if (r == leavingRow)
						etaMatrix[r, leavingRow] = 1 / dxi;
					else
						etaMatrix[r, leavingRow] = -stepDirection[r, 0] / dxi;
				}
				etaFile.Add(etaMatrix);

				float tempF = cNonbasic[enteringCol, 0];
				cNonbasic[enteringCol, 0] = cBasic[leavingRow, 0];
				cBasic[leavingRow, 0] = tempF;

				int tempI = nonbasicVars[enteringCol];
				nonbasicVars[enteringCol] = basicVars[leavingRow];
				basicVars[leavingRow] = tempI;

				if (printControl.variableLists) {
					Console.Write("Basic:    ");
					for (int i = 0; i < m; ++i)
						Console.Write(basicVars[i] + " ");
					Console.Write("\nNonbasic: ");
					for (int i = 0; i < n; ++i)
						Console.Write(nonbasicVars[i] + " ");
					Console.WriteLine();
				}
				if (printControl.iterationCount) Console.WriteLine("\nCompleted iteration {0}", iterationCount++);
				if (printControl.iterationSeparator) Console.WriteLine("\n\n");
			}
			Console.ReadLine();
		}
	}
}
