#include "asynplexcontrol.h"
#include "print.h"
#include "communication.h"
#include <malloc_align.h>
#include <free_align.h>
#include "common/types.h"

extern spe_program_handle_t speProgramHandle;

static AsynplexControl * instance; // Hack to give the static RPC functions access to the AsynplexControl object (of which there is only one)

// Input: none
// Output: A two element int array, containing m and n (will not be freed, but we can tolerate an eight byte leak; this function is only called once)
void AsynplexControl::getDimensions(cellmsg_rpc_data * rpc) {
	rpc->buffer = _malloc_align(4 * sizeof(int), ALIGN_BUS_LOG2);
	((int*)rpc->buffer)[0] = instance->m;
	((int*)rpc->buffer)[1] = instance->n;
	((int*)rpc->buffer)[2] = instance->numNonzeroes;
	rpc->numbytes = 4 * sizeof(int);
}

void AsynplexControl::getIntArrays(cellmsg_rpc_data * rpc) {
	rpc->buffer = instance->intArrays;
	rpc->numbytes = instance->intBytes;
}

void AsynplexControl::getTypeArrays(cellmsg_rpc_data * rpc) {
	rpc->buffer = instance->typeArrays;
	rpc->numbytes = instance->typeBytes;
}

void AsynplexControl::runAsynplex(int numIterationProcesses, int m, int n, SparseMatrix & A, TYPE * b, TYPE * c, TYPE f) {
	instance = this;
	this->m = m;
	this->n = n;
	int N = m + n;
	this->numNonzeroes = A.numNonzeroes;
	this->numIterationProcesses = numIterationProcesses;
	int intCount = numNonzeroes + (N + 1);		// A.rowIndices, A.colPos
	int typeCount = 1 + m + N + numNonzeroes;	// f, b, c, A.values
	this->intBytes = ROUND_UP_MULTIPLE(intCount * sizeof(int), ALIGN_BUS_WIDTH);
	this->typeBytes = ROUND_UP_MULTIPLE(typeCount * sizeof(TYPE), ALIGN_BUS_WIDTH);
	this->intArrays = (int*)_malloc_align(intBytes, ALIGN_BUS_LOG2);
	this->typeArrays = (TYPE*)_malloc_align(typeBytes, ALIGN_BUS_LOG2);
	memcpy(this->intArrays, A.rowIndices, numNonzeroes * sizeof(int));
	memcpy(this->intArrays + numNonzeroes, A.colPos, (N + 1) * sizeof(int));
	this->typeArrays[0] = f;
	memcpy(this->typeArrays + 1, b, m * sizeof(TYPE));
	memcpy(this->typeArrays + 1 + m, c, N * sizeof(TYPE));
	memcpy(this->typeArrays + 1 + m + N, A.values, numNonzeroes * sizeof(TYPE));
	initPrintFromProcess();
	initCommunication();
	setupInvertProcessor(A, numIterationProcesses);
	pthread_create(&invertProcessor, NULL, runInvertProcessor, NULL);
	
	printf("m: %d, n: %d, nz: %d\n", m, n, A.numNonzeroes);fflush(stdout);
	cellmsg_provide_rpc(sendRpcMessage);
	cellmsg_provide_rpc(sendRpcMessageBatch);
	cellmsg_provide_rpc(receiveRpcMessage);
	cellmsg_provide_rpc(getDimensions);
	cellmsg_provide_rpc(getIntArrays);
	cellmsg_provide_rpc(getTypeArrays);
	cellmsg_init(0, NULL);
	cellmsg_run(&speProgramHandle, 0, NULL); // Start the SPE program running on all SPEs and wait until all SPEs invoke MPI_Finalize()
	cellmsg_finalize();
}

void AsynplexControl::waitForCompletion() {
	pthread_join(invertProcessor, NULL);
}
