#include <stdio.h>
#include "tree.h"
#include "myalloc.h"

#define TRUE  1
#define FALSE 0

static void killnode( TreeNode *node )
{
    if (node->left  != NULL) killnode(node->left);;
    if (node->right != NULL) killnode(node->right);;
    if (node != NULL) FREE(node);
}

void killtree( Tree * tree )
{
    if (tree->root != NULL) killnode(tree->root);
    tree->root = NULL;
    tree->curnode = NULL;
}

void addtree( Tree * tree, int data )
{
    TreeNode *node = tree->root;
    TreeNode *parent;

    if (tree->root == NULL) {
		MALLOC( tree->root, 1, TreeNode );
		tree->root->data   = data;
		tree->root->parent = NULL;
		tree->root->left   = NULL;
		tree->root->right  = NULL;
		return;
    }
    while ( node != NULL ) {
	     if ( node->data > data ) { parent = node; node = node->left; }
	else if ( node->data < data ) { parent = node; node = node->right; }
	else                          { return; }	/* already there */
    }

    MALLOC( node, 1, TreeNode );
    node->data   = data;
    node->parent = parent;
    node->left   = NULL;
    node->right  = NULL;
    if (parent->data > data) { parent->left  = node; }
    else                     { parent->right = node; }
}

void deltree( Tree * tree, int data )
{
    TreeNode *node   = tree->root;
    TreeNode *parent = NULL;
    TreeNode *node1, *node2;

    while ( node != NULL && node->data != data ) {
	     if ( node->data > data ) { node = node->left; }
	else if ( node->data < data ) { node = node->right; }
    }

    if (node == NULL) { return; }  /* not there anyway */

    parent = node->parent;

    if (node->right == NULL) {
	node1 = node->left;
	if (node1 != NULL) { node1->parent = parent; }
    } else if (node->left == NULL) {
	node1 = node->right;
	node1->parent = parent;
    } else {
	node1 = node->left;
        for ( node2 = node1; node2->right != NULL; node2 = node2->right ) { }
        node2->right = node->right;
	node->right->parent = node2;
    
        node->left->parent = parent;
    }
    if (parent == NULL) {
	tree->root = node1;
    } else {
        if (parent->data > data) { parent->left  = node1; }
	                    else { parent->right = node1; }
    }

    FREE( node );
}

int getfirst( Tree * tree )
{
    TreeNode *node;
    TreeNode *parent;

    if (tree->root == NULL) { /* printf("getfirst: empty tree \n"); */ return -1; }

    for (node=tree->root; node!=NULL; parent=node, node=node->left) { }

    tree->curnode=parent;

    return tree->curnode->data;
}

int getnext( Tree * tree )
{
    TreeNode *node;
    TreeNode *par;

    if (tree->curnode==NULL) { /*printf("getnext: no current node \n");*/ return -1; }

    if (tree->curnode->right!=NULL) {
        for (node=tree->curnode->right; node!=NULL; par=node, node=node->left) { }
        tree->curnode = par;
        return tree->curnode->data;
    }
    for (node=tree->curnode->parent; node!=NULL; node=node->parent) {
	 if ( node->data > tree->curnode->data ) break;
    }
    tree->curnode = node;
    if (tree->curnode!=NULL) {
    	return tree->curnode->data;
    } else {
    	return -1;		/* no more */
    }
}

int getlast( Tree * tree )
{
    TreeNode *node;
    TreeNode *parent;

    if (tree->root == NULL) { /* printf("getlast: empty tree \n"); */ return -1; }

    for (node=tree->root; node!=NULL; parent=node, node=node->right) { }

    tree->curnode=parent;

    return tree->curnode->data;
}

int getprev( Tree * tree )
{
    TreeNode *node;
    TreeNode *par;

    if (tree->curnode==NULL) { /*printf("getnext: no current node \n");*/ return -1; }

    if (tree->curnode->left!=NULL) {
        for (node=tree->curnode->left; node!=NULL; par=node, node=node->right) { }
        tree->curnode = par;
        return tree->curnode->data;
    }
    for (node=tree->curnode->parent; node!=NULL; node=node->parent) {
	 if ( node->data < tree->curnode->data ) break;
    }
    tree->curnode = node;
    if (tree->curnode!=NULL) {
    	return tree->curnode->data;
    } else {
    	return -1;		/* no more */
    }
}

static void printnode( TreeNode *node )
{
    int parent, left, right;

    if (node != NULL) {
        if (node->parent!=NULL) {parent=node->parent->data;} else {parent = -1;}
        if (node->left  !=NULL) {left  =node->left  ->data;} else {left   = -1;}
        if (node->right !=NULL) {right =node->right ->data;} else {right  = -1;}
	printnode( node->left );
	printf("    %4d %4d %4d %4d\n", node->data, parent, left, right);
	printnode( node->right );
    }
}

void printtree( Tree * tree )
{
    if (tree->root != NULL) printf("root node is %4d\n", tree->root->data);
    printf("  data parent left right\n");
    printnode( tree->root );
}
