#include "../common/message.h"
#include "../common/sparse.h"
#include "communication.h"
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <assert.h>
#include <malloc_align.h>
#include <free_align.h>
#include "print.h"
#include "common/SafeVector.h"
#include "common/tags.h"
#include "common/types.h"

Message createMessage(const char * sender, const char * receiver, const char * tag, void * payload) {
	Message message;
	message.sender = sender;
	message.receiver = receiver;
	message.tag = tag;
	message.payload = payload;
	return message;
}

static SafeVector<RpcMessage> messages;
static pthread_mutex_t queueLock;
static int printSends = 1;
static int printReceives = 1;

void initCommunication() {
	pthread_mutex_init(&queueLock, NULL);
}

void sendMessage(RpcMessage message) {
	pthread_mutex_lock(&queueLock);
	messages.push_back(message);
	pthread_mutex_unlock(&queueLock);
	if (printSends) {
		printFromProcess("COMM", "%d sends to %d (tag: %d, payload %p) (queue: %d)", message.sender, message.receiver, message.tag, message.payload, messages.size());
	}
}
/*
void sendAll(Message * messagesToBeSent, int numMessages) {
	int i;
	pthread_mutex_lock(&queueLock);
	for (i = 0; i < numMessages; ++i) {
		messages.push_back(messagesToBeSent[i]);
		if (printSends) {
			printFromProcess("COMM", "%s sends to %s (tag: '%s', payload %p) (queue: %d)", messagesToBeSent[i].sender, messagesToBeSent[i].receiver, messagesToBeSent[i].tag, messagesToBeSent[i].payload, messages.size());
		}
	}
	pthread_mutex_unlock(&queueLock);
}
*/
// sender < 0 means "receive from any iteration process"
// tag < 0 means "any tag"
RpcMessage receiveMessageWithTag(int receiver, int sender, int tag) {
	RpcMessage msg;
	int i;
	pthread_mutex_lock(&queueLock);
	for (i = 0; i < messages.size(); ++i) {
		msg = messages[i];
		if (msg.receiver == receiver && ((sender < 0 && msg.sender >= RANK_OF_FIRST_ITER_PROC) || msg.sender == sender) && (tag < 0 || msg.tag == tag)) {
			messages.erase(messages.begin() + i);
			if (printReceives) {
				printFromProcess("COMM", "%d receives from %d (tag: %d, payload: %p) (queue: %d)", msg.receiver, msg.sender, msg.tag, msg.payload, messages.size());
			}
			pthread_mutex_unlock(&queueLock);
			return msg;
		}
	}
	pthread_mutex_unlock(&queueLock);
	msg.sender = -1;
	return msg;
}

void printMessages() {
	int i;
	pthread_mutex_lock(&queueLock);
	printFromProcess("COMM", "Remaining messages:");
	for (i = 0; i < messages.size(); ++i) {
		printFromProcess("COMM", "%s %s %s", messages[i].sender, messages[i].receiver, messages[i].tag);
	}
	fflush(stdout);
	pthread_mutex_unlock(&queueLock);
}

void sendRpcMessage(cellmsg_rpc_data * rpc) {
	RpcMessage message;
	int * buffer = (int*)rpc->buffer;
	message.sender = buffer[0];
	message.receiver = buffer[1];
	message.tag = buffer[2];
	message.payloadSize = buffer[3];
	assert(message.payloadSize <= rpc->numbytes);
	message.payload = _malloc_align(message.payloadSize, ALIGN_BUS_LOG2);
	memcpy(message.payload, buffer, message.payloadSize);
	sendMessage(message);
	rpc->numbytes = 0; // No RPC output
}

void sendRpcMessageBatch(cellmsg_rpc_data * rpc) { //TODO
	assert(0);
}

//static void * previouslyReceivedBuffer = NULL;

void receiveRpcMessage(cellmsg_rpc_data * rpc) {
	int * buffer = (int*)rpc->buffer;
	int sender = buffer[0];
	int receiver = buffer[1];
	int tag = buffer[2];
	RpcMessage message = receiveMessageWithTag(receiver, sender, tag);
	if (message.sender < 0) { // No suitable message was found
		rpc->numbytes = sizeof(int);
		buffer[0] = -1; // Sender < 0 signifies no message
	}
	else {
		rpc->buffer = message.payload; //TODO: Must be freed somewhere
		rpc->numbytes = message.payloadSize;
	}
}
