#include "communication.h"
#include "invertprocessor.h"
#include "common/types.h" // For the FactoredInverse structure TODO: restructure
#include "../common/sparse.h"
#include "common/util.h"
#include "payloadtypes.h"
#include "../common/myalloc.h"
#include "../common/macros.h"
#include "../common/heap.h"
#include "../common/linalg.h"
#include "print.h"
#include <stdio.h>
#include <pthread.h>
#include <assert.h>
#include <vector>
using namespace std;

static const char * NAME = "V";

static SparseMatrix A;
static int * basicVariables;
static int * nonbasicVariables;
static int m, n, N;
static int verbose = 0;
static SparseMatrix L = {0, 0, 0, NULL, NULL, NULL};
static SparseMatrix LT = {0, 0, 0, NULL, NULL, NULL};
static SparseMatrix U = {0, 0, 0, NULL, NULL, NULL};
static SparseMatrix UT = {0, 0, 0, NULL, NULL, NULL};

typedef struct {    /* nonzero entry */
	TYPE value;  /* value */
	int rowIndex;  /* row index */
} ValueAndRowIndex;

static SafeVector<int> colPerm, invColPerm, rowPerm, invRowPerm;
static SafeVector<TYPE> diagU;

static double cumtime = 0.0; //TODO: This one was used both by the functions that were moved to invertprocessor.c and to interationprocess.c

static int numIterationProcesses;
static char ** iterationProcessNames;

#define E_N		200
#define E_NZ	20000
#define MD		1
#define EPSNUM	1.0e-9

static int		rank;

static int receivedBasisNumber = 0;
static int lastReinversionBasis = 0;

static pthread_mutex_t finishLock;
static int _isProgramFinished = 0;

static void invertProcessor_lufac();
static void invertProcessor_broadcastInverse();

static void send(const char * receiver, const char * tag, void * payload) {
	Message msg = {NAME, receiver, tag, payload};
	//sendMessage(msg);
}

static Message receiveFromAnyone() {
	Message message;
	return message;
	//return receiveMessageFromAnyone(NAME);
}

//http://ou800doc.sco.com/cgi-bin/info2html?(gcc.info)Macro%2520Varargs&lang=en
#define print(formatString, ...) printFromProcess(NAME, formatString , ## __VA_ARGS__)

int isProgramFinished() {
	int result;
	pthread_mutex_lock(&finishLock);
	result = _isProgramFinished;
	pthread_mutex_unlock(&finishLock);
	return result;
}

int announceProgramFinished() {
	int oldValue;
	pthread_mutex_lock(&finishLock);
	oldValue = _isProgramFinished;
	_isProgramFinished = 1;
	pthread_mutex_unlock(&finishLock);
	return oldValue;
}

void setupInvertProcessor(SparseMatrix _A, int _numIterationProcesses) {
	int i;
	numIterationProcesses = _numIterationProcesses;
	iterationProcessNames = (char**)malloc(numIterationProcesses * sizeof(char *));
	for (i = 0; i < numIterationProcesses; ++i) {
		iterationProcessNames[i] = (char*)malloc(16);
		sprintf(iterationProcessNames[i], "I%d", i);
	}
	copySparseMatrix(_A, &A);
	m = A.rows;
	N = A.cols;
	n = N - m;
	basicVariables = (int*)malloc(m * sizeof(int));
	for (i = 0; i < m; ++i) {
		basicVariables[i] = n + i;
	}
	nonbasicVariables = (int*)malloc(n * sizeof(int));
	for (i = 0; i < n; ++i) {
		nonbasicVariables[i] = i;
	}
	pthread_mutex_init(&finishLock, NULL);
}

void * runInvertProcessor(void * parameters) {
	pthread_exit(NULL);
	return NULL;
	Message message;
	InvProcPivotPayload * pivot;
	int temp;
	print("Starting");
	seedRandomGeneratorForThisThread();
	
	while (!isProgramFinished()) {
		while ((message = receiveFromAnyone()).sender || receivedBasisNumber == lastReinversionBasis) {
			if (isProgramFinished()) {
				print("Terminating");
				pthread_exit(NULL);
				return NULL;
			}
			if (!message.sender)
				continue;
			assert(!strcmp(message.tag, "I8->V1"));
			pivot = (InvProcPivotPayload*)message.payload;
			//print("Received pivot: l/e: %d/%d", pivot->leavingRow, pivot->enteringCol);
			temp = basicVariables[pivot->leavingRow];
			basicVariables[pivot->leavingRow] = nonbasicVariables[pivot->enteringCol];
			nonbasicVariables[pivot->enteringCol] = temp;
			++receivedBasisNumber;
			free(pivot);
		}
		if (isProgramFinished()) {
			print("Terminating");
			pthread_exit(NULL);
			return NULL;
		}
	
		print("Starting inverse calculation for basis %d", receivedBasisNumber);
		invertProcessor_lufac();
		lastReinversionBasis = receivedBasisNumber;
		print("Inverse for basis %d done; broadcasting", lastReinversionBasis);
		invertProcessor_broadcastInverse();
		/*
		printRawSparseMatrix(L);
		printRawSparseMatrix(U);
		for (i = 0; i < m; ++i)
			printf("%f ", diagU[i]);
		printf("\n");
		*/
	}
	
	print("Terminating");
	pthread_exit(NULL);
	return NULL; // To prevent a compiler warning
}

static void invertProcessor_broadcastInverse() {
	int i;
	FactoredInverse * factoredInverse;
	Message * messages = new Message[numIterationProcesses];
	for (i = 0; i < numIterationProcesses; ++i) {
		//TODO: Do we want each process to get its own copy? If not, how do we delete the common copy after all processes are done with it?
		//TODO: Each ItProc must free BOTH the matrices/vectors AND the FactoredInverse object
		factoredInverse = new FactoredInverse;
		copySparseMatrix(L, &factoredInverse->L);
		copySparseMatrix(LT, &factoredInverse->LT);
		copySparseMatrix(U, &factoredInverse->U);
		copySparseMatrix(UT, &factoredInverse->UT);
		factoredInverse->colPerm = colPerm;
		factoredInverse->invColPerm = invColPerm;
		factoredInverse->rowPerm = rowPerm;
		factoredInverse->invRowPerm = invRowPerm;
		factoredInverse->diagU = diagU;
		factoredInverse->rank = rank;
		factoredInverse->basisNumber = lastReinversionBasis;
		messages[i].sender = NAME;
		messages[i].receiver = iterationProcessNames[i];
		messages[i].tag = "V2->I1";
		messages[i].payload = factoredInverse;
	}
	//TODO sendAll(messages, numIterationProcesses);
	delete [] messages;
}

static void invertProcessor_lufac() {
	int kk, kkk, tag, rowdeg, coldeg, row, col, row2, col2;

	int i, j, k, cnt, lnz, unz, lnzbnd, unzbnd, okey, deg, heapnum, cur,
			method = MD;

	int *degB = NULL, *degBt = NULL, *hkey = NULL, *heap = NULL, *iheap = NULL,
			*iwork = NULL, *iwork2 = NULL;

	ValueAndRowIndex tempB, **B = NULL, **Bt = NULL;

	TYPE narth;

	double starttime, endtime;

	starttime = (double) clock();

	/*---------------------------------------------------------+
	 | allocate space for perm and iperm.                      */
	colPerm.resize(m);
	invColPerm.resize(m);
	rowPerm.resize(m);
	invRowPerm.resize(m);
	/*---------------------------------------------------------+
	 | allocate space for work arrays.                         */

	MALLOC( degB, m, int );
	MALLOC( degBt, m, int );
	MALLOC( hkey, m, int );
	MALLOC( heap, m, int );
	MALLOC( iheap, m, int );
	MALLOC( iwork, m, int );
	MALLOC( iwork2, m, int );

	heap--; /* so that indexing starts from 1 */

	/*---------------------------------------------------------+
	 | calculate degrees in B and Bt                           */

	for (i = 0; i < m; i++) {
		degBt[i] = 0;
	}
	for (i = 0; i < m; i++) {
		degB[i] = A.colPos[basicVariables[i] + 1] - A.colPos[basicVariables[i]];
		for (k = A.colPos[basicVariables[i]]; k < A.colPos[basicVariables[i] + 1]; k++) {
			degBt[A.rowIndices[k]]++;
		}
	}
	
	/*---------------------------------------------------------+
	 | calculate initial estimate of number of nonzeros in      |
	 | L and UT.values                                                */

	lnzbnd = 0;
	for (i = 0; i < m; i++)
		lnzbnd += degB[i];
	lnzbnd = lnzbnd / 2;
	unzbnd = lnzbnd;

	/*---------------------------------------------------------+
	 | allocate enough space to store L and UT.values                  |
	 | (without any fillin)                                    */

	if (L.colPos == NULL) {
		MALLOC( L.colPos, m+1, int );
	} else {
		REALLOC( L.colPos, m+1, int );
	}
	if (L.rowIndices == NULL) {
		MALLOC( L.rowIndices, lnzbnd, int );
	} else {
		REALLOC( L.rowIndices, lnzbnd, int );
	}
	if (L.values == NULL) {
		MALLOC( L.values, lnzbnd, TYPE );
	} else {
		REALLOC( L.values, lnzbnd, TYPE );
	}
	if (UT.colPos == NULL) {
		MALLOC( UT.colPos, m+1, int );
	} else {
		REALLOC( UT.colPos, m+1, int );
	}
	if (UT.rowIndices == NULL) {
		MALLOC( UT.rowIndices, unzbnd, int );
	} else {
		REALLOC( UT.rowIndices, unzbnd, int );
	}
	if (UT.values == NULL) {
		MALLOC( UT.values, unzbnd, TYPE );
	} else {
		REALLOC( UT.values, unzbnd, TYPE );
	}
	diagU.resize(m);
	
	MALLOC( B, m, ValueAndRowIndex * );
	MALLOC( Bt, m, ValueAndRowIndex * );
	for (i = 0; i < m; i++) {
		B[i] = NULL;
		Bt[i] = NULL;
		MALLOC( B[i], degB[i], ValueAndRowIndex );
		MALLOC( Bt[i], degBt[i], ValueAndRowIndex );
	}

	/*---------------------------------------------------------+
	 | initialize B and Bt                                     */

	for (i = 0; i < m; i++) {
		iwork[i] = 0;
	}
	for (j = 0; j < m; j++) {
		kkk = 0;
		for (k = A.colPos[basicVariables[j]]; k < A.colPos[basicVariables[j] + 1]; k++) {
			row = A.rowIndices[k];
			kk = iwork[row];
			B[j][kkk].rowIndex = row;
			B[j][kkk].value = A.values[k];
			Bt[row][kk].rowIndex = j;
			Bt[row][kk].value = A.values[k];
			iwork[row]++;
			kkk++;
		}
	}

	/*---------------------------------------------------------+
	 | miscellaneous initializations.                          */

	for (i = 0; i < m; i++) {
		invColPerm.at(i) = -1;
		invRowPerm.at(i) = -1;
		iwork[i] = 0;
		iwork2[i] = -1;
	}

	rank = m;
	tag = 0;
	lnz = 0;
	unz = 0;
	L.colPos[0] = 0;
	UT.colPos[0] = 0;

	/*---------------------------------------------------------+
	 | hkey encodes the tie-breaking rule - currently the rule  |
	 | is somewhat random.  to make it first occuring minimum,  |
	 | change the formula to:                                   |
	 |       hkey[node] = degree[node]*m + node;                |
	 | warning: with this definition of hkey, there is the      |
	 | possibility of integer overflow on moderately large      |
	 | problems.                                                |
	 |                                                         */

	for (j = 0; j < m; j++) {
		if (method == MD)
			hkey[j] = degB[j];
		else
			hkey[j] = j;

		if (hkey[j] == 0)
			hkey[j] = m + 1;
	}

	/*---------------------------------------------------------+
	 | set up heap structure for quickly accessing minimum.    */

	heapnum = m;
	for (j = m - 1; j >= 0; j--) {
		cur = j + 1;
		iheap[j] = cur;
		heap[cur] = j;
		hfall(heapnum, hkey, iheap, heap, cur);
	}

	/*---------------------------------------------------------+
	 | the min degree ordering loop                            */

	for (i = 0; i < m; i++) {

		/*------------------------------------------------+
		 |  select column with min column degree          */

		again: col = heap[1];
		coldeg = degB[col];

		if (coldeg == 0) {
			printf("singular matrix. rank deficiency = %d\n", m - i);
			rank = i;
			goto end;
		}

		/*------------------------------------------------+
		 |  select pivot element from this column by       |
		 |  choosing nonzero whose row is of minimal       |
		 |  degree                                        */

		rowdeg = m + 1;
		for (k = 0; k < coldeg; k++) {
			if (degBt[B[col][k].rowIndex] < rowdeg && ABS( B[col][k].value ) > EPSNUM) {
				row = B[col][k].rowIndex;
				rowdeg = degBt[row];
			}
		}
		if (rowdeg == m + 1) {
			hkey[col] = m + 2;
			hfall(heapnum, hkey, iheap, heap, iheap[col]);
			if (hkey[heap[1]] == m + 2) {
				printf("singular matrix. rank deficiency = %d\n", m - i);
				rank = i;
				goto end;
			} else {
				goto again;
			}
		}

		/*------------------------------------------------+
		 |  update permutation information                */

		colPerm.at(i) = col;
		invColPerm.at(col) = i;

		rowPerm.at(i) = row;
		invRowPerm.at(row) = i;

		/*------------------------------------------------+
		 |  reallocate space for L, L.rowIndices, UT.values, and UT.rowIndices as     |
		 |        necessary.                               |
		 |                                                 |
		 |  lnz stores the number of nonzeros in L so far  |
		 |  lnzbnd is an estimate of how many will be in L |
		 |  unz stores the number of nonzeros in U so far  |
		 |  unzbnd is an estimate of how many will be in U*/

		cnt = lnz + coldeg - 1 + coldeg * rowdeg / 2;
		if (cnt > lnzbnd) {
			lnzbnd = cnt;
			REALLOC( L.values, lnzbnd, TYPE );
			REALLOC( L.rowIndices, lnzbnd, int );
		}

		cnt = unz + rowdeg - 1 + coldeg * rowdeg / 2;
		if (cnt > unzbnd) {
			unzbnd = cnt;
			REALLOC( UT.values, unzbnd, TYPE );
			REALLOC( UT.rowIndices, unzbnd, int );
		}

		/*------------------------------------------------+
		 |  copy pivot column into L and pivot row into    |
		 |  Ut.                                           */

		L.colPos[i + 1] = L.colPos[i] + coldeg - 1;

		for (k = 0; k < coldeg; k++) {
			if (B[col][k].rowIndex != row) {
				L.rowIndices[lnz] = B[col][k].rowIndex;
				L.values[lnz] = B[col][k].value;
				lnz++;
			}
		}

		UT.colPos[i + 1] = UT.colPos[i] + rowdeg - 1;

		for (k = 0; k < rowdeg; k++) {
			if (Bt[row][k].rowIndex != col) {
				UT.rowIndices[unz] = Bt[row][k].rowIndex;
				UT.values[unz] = Bt[row][k].value;
				unz++;
			} else {
				diagU[i] = Bt[row][k].value;
			}
		}

		/*------------------------------------------------+
		 |  remove eliminated elements from B and Bt      */

		for (k = 0; k < coldeg; k++) {
			row2 = B[col][k].rowIndex;
			degBt[row2]--;
			for (kk = 0; Bt[row2][kk].rowIndex != col; kk++)
				;

			tempB = Bt[row2][degBt[row2]];
			Bt[row2][degBt[row2]] = Bt[row2][kk];
			Bt[row2][kk] = tempB;
		}

		for (k = 0; k < rowdeg; k++) {
			col2 = Bt[row][k].rowIndex;
			degB[col2]--;
			for (kk = 0; B[col2][kk].rowIndex != row; kk++)
				;

			tempB = B[col2][degB[col2]];
			B[col2][degB[col2]] = B[col2][kk];
			B[col2][kk] = tempB;
		}
		degB[col] = 0;
		degBt[row] = 0;

		/*------------------------------------------------+
		 |  update heap                                   */

		okey = hkey[col];
		heap[1] = heap[heapnum];
		iheap[heap[1]] = 1;
		heapnum--;
		if (okey < hkey[heap[1]])
			hfall(heapnum, hkey, iheap, heap, 1);

		/*------------------------------------------------+
		 |  generate fillin and update elements           */

		for (k = L.colPos[i]; k < L.colPos[i + 1]; k++) {
			row2 = L.rowIndices[k];
			tag++;
			for (kk = 0; kk < degBt[row2]; kk++) {
				iwork[Bt[row2][kk].rowIndex] = tag; /* tag these columns */
				iwork2[Bt[row2][kk].rowIndex] = kk; /* say where they are */
			}
			for (kk = UT.colPos[i]; kk < UT.colPos[i + 1]; kk++) {
				col2 = UT.rowIndices[kk];
				if (iwork[col2] == tag) {
					Bt[row2][iwork2[col2]].value -= L.values[k] * UT.values[kk]
							/ diagU[i];
				} else {
					deg = degBt[row2];
					REALLOC( Bt[row2], deg+1, ValueAndRowIndex );
					Bt[row2][deg].rowIndex = col2;
					Bt[row2][deg].value = -L.values[k] * UT.values[kk] / diagU[i];
					degBt[row2]++;
				}
			}
		}

		for (k = UT.colPos[i]; k < UT.colPos[i + 1]; k++) {
			col2 = UT.rowIndices[k];
			tag++;
			for (kk = 0; kk < degB[col2]; kk++) {
				iwork[B[col2][kk].rowIndex] = tag; /* tag these rows */
				iwork2[B[col2][kk].rowIndex] = kk; /* say where they are */
			}
			for (kk = L.colPos[i]; kk < L.colPos[i + 1]; kk++) {
				row2 = L.rowIndices[kk];
				if (iwork[row2] == tag) {
					B[col2][iwork2[row2]].value -= L.values[kk] * UT.values[k]
							/ diagU[i];
				} else {
					deg = degB[col2];
					REALLOC( B[col2], deg+1, ValueAndRowIndex );
					B[col2][deg].rowIndex = row2;
					B[col2][deg].value = -L.values[kk] * UT.values[k] / diagU[i];
					degB[col2]++;
				}
			}
		}

		/*------------------------------------------------+
		 |  adjust heap                                   */

		for (k = UT.colPos[i]; k < UT.colPos[i + 1]; k++) {
			col2 = UT.rowIndices[k];
			if (method == MD) {
				hkey[col2] = degB[col2];
			} else {
				hkey[col2] = col2;
			}
			if (hkey[col2] == 0)
				hkey[col2] = m + 1;
			hrise(hkey, iheap, heap, iheap[col2]);
			hfall(heapnum, hkey, iheap, heap, iheap[col2]);
		}

		/*------------------------------------------------+
		 |  free space no longer needed                   */

		/*
		 FREE(B[col]);  FREE(Bt[row]);
		 */
	}
	end:
	/*------------------------------------------------+
	 |  process dependent rows/cols                   */

	i = rank;
	for (col = 0; col < m; col++) {
		if (invColPerm.at(col) == -1) {
			colPerm.at(i) = col;
			invColPerm.at(col) = i;
			i++;
		}
	}

	i = rank;
	for (row = 0; row < m; row++) {
		if (invRowPerm.at(row) == -1) {
			rowPerm.at(i) = row;
			invRowPerm.at(row) = i;
			i++;
		}
	}

	for (i = rank; i < m; i++) {
		L.colPos[i + 1] = L.colPos[i];
		UT.colPos[i + 1] = UT.colPos[i];
		diagU[i] = 0.0;
	}

	/*------------------------------------------------+
	 |  free up space                                 */

	heap++;
	for (i = 0; i < m; i++) {
		FREE( B[i] );
		FREE( Bt[i] );
	}
	FREE(degB);
	FREE(degBt);
	FREE(hkey);
	FREE(heap);
	FREE(iheap);
	FREE(iwork);
	FREE(iwork2);
	FREE(B);
	FREE(Bt);

	/*------------------------------------------------+
	 |  update "i" arrays to new indices              */

	for (k = 0; k < L.colPos[m]; k++)
		L.rowIndices[k] = invRowPerm.at(L.rowIndices[k]);
	for (k = 0; k < UT.colPos[m]; k++)
		UT.rowIndices[k] = invColPerm.at(UT.rowIndices[k]);

	/*------------------------------------------------+
	 |  divide each column of L by diagonal           */

	for (i = 0; i < m; i++) {
		for (k = L.colPos[i]; k < L.colPos[i + 1]; k++) {
			L.values[k] /= diagU[i];
		}
	}

	/*---------------------------------------------------------+
	 | calculate and print statistics.                         */

	narth = 0.0e0;
	for (i = 0; i < m; i++) {
		k = L.colPos[i + 1] - L.colPos[i];
		narth += (TYPE) k * k;
		k = UT.colPos[i + 1] - UT.colPos[i];
		narth += (TYPE) k * k;
	}
	narth = narth + 3* L .colPos[m] + 3* UT .colPos[m] + 2* m ;

	lnz = L.colPos[m];
	unz = UT.colPos[m];
	if (verbose) {
		printf("%9d   %9d %15.0f ", lnz, unz, narth);
		fflush(stdout);
	}

	if (LT.values == NULL) {
		MALLOC( LT.values, lnz, TYPE);
	} else {
		REALLOC( LT.values, lnz, TYPE);
	}
	if (LT.rowIndices == NULL) {
		MALLOC( LT.rowIndices, lnz, int);
	} else {
		REALLOC( LT.rowIndices, lnz, int);
	}
	if (LT.colPos == NULL) {
		MALLOC( LT.colPos, m+1, int);
	} else {
		REALLOC( LT.colPos, m+1, int);
	}

	if (U.values == NULL) {
		MALLOC( U.values, unz, TYPE);
	} else {
		REALLOC( U.values, unz, TYPE);
	}
	if (U.rowIndices == NULL) {
		MALLOC( U.rowIndices, unz, int);
	} else {
		REALLOC( U.rowIndices, unz, int);
	}
	if (U.colPos == NULL) {
		MALLOC( U.colPos, m+1, int);
	} else {
		REALLOC( U.colPos, m+1, int);
	}
	
	L.rows = LT.rows = U.rows = UT.rows = L.cols = LT.cols = U.cols = UT.cols = m;
	L.numNonzeroes = lnz;
	LT.numNonzeroes = lnz;
	U.numNonzeroes = unz;
	UT.numNonzeroes = unz;
	transposeMatrix(m, m, L.colPos, L.rowIndices, L.values, LT.colPos, LT.rowIndices, LT.values);
	transposeMatrix(m, m, UT.colPos, UT.rowIndices, UT.values, U.colPos, U.rowIndices, U.values);
	
	endtime = (double) clock();
	cumtime += endtime - starttime;
}
