#include "ColumnSelectionManager.h"
#include <assert.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <libmisc.h>
#include "rpc.h"
#include "common/myalloc.h"
using namespace std;

#define NAME "C"

#define RANK RANK_CS

static char * VARIABLE_STATUS_NAMES[] = {"OTHER_SIDE", "UNSELECTED", "SELECTED", "REJECTED", "UNDEFINED"};

//#define print(formatString, ...) printFromProcess(NAME, formatString , ## __VA_ARGS__)

void ColumnSelectionManager::setup(int m, int n) {
	int i;
	this->m = m;
	this->n = n;
	printf("ColSel: %d %d\n", m, n);fflush(stdout);
	kc = 0;
	MALLOC(variables, m + n, VariableInfo);
	//TODO: This must be switched between the phases
	for (i = 0; i < n; ++i) { // nonbasic
		variables[i].status = OTHER_SIDE;
		variables[i].basisWhereStatusChanged = -1;
	}
	for (i = 0; i < m; ++i) { // basic
		variables[n + i].status = UNSELECTED;
		variables[n + i].basisWhereStatusChanged = -1;
	}
}

bool ColumnSelectionManager::installAttractiveCandidates(const SafeVector<int> & newCandidates, int basisNumber) {
	assert(newCandidates.size() >= 1);
	//print("received %d candidates from %s, which is at basis %d; currently at basis %d", newCandidates.size(), message.sender, msg.basisNumber, kc);
	if (basisNumber > kc) {
		int installCount = 0;
		attractive.clear();
		for (int i = 0; i < newCandidates.size(); ++i) {
			int cand = newCandidates[i];
			if (!((variables[cand].status == SELECTED || variables[cand].status == REJECTED) && variables[cand].basisWhereStatusChanged >= basisNumber)) {
				++installCount;
				attractive.push_back(cand);
				variables[cand].status = UNSELECTED;
				variables[cand].basisWhereStatusChanged = basisNumber;
			}
			else {
				//print("rejected x%d because it got %s at basis %d", cand, VARIABLE_STATUS_NAMES[variables[cand].status], variables[cand].basisWhereStatusChanged);
			}
		}
		//print("Installed %d candidates", installCount);
		kc = basisNumber;
		return true;
	}
	else {
		//print("Rejected the candidates");
		return false;
	}
}

void ColumnSelectionManager::run() {
	/*int * buffer = (int*)malloc_align(8 * sizeof(int), ALIGN_BUS_LOG2);
	buffer[0] = RANK; // Sender
	buffer[1] = RANK_BC; // Receiver
	buffer[2] = 23; // Tag
	buffer[3] = 8 * sizeof(int); // Size
	buffer[4] = 42; // Payload
	printf("Sending message\n");fflush(stdout);
	cellmsg_rpc(sendRpcMessage, 
				buffer, buffer[3], CML_BYTE_SWAP_NOT_NEEDED, 
				NULL, 0, CML_BYTE_SWAP_NOT_NEEDED);
	printf("Message sent\n");fflush(stdout);*/
	/*
	Message message;
	int var;
	int * varPayload;
	QueuedVariableRequest req;
	print("Column selection manager starting");
	
	while (!isProgramFinished()) {
		while (queuedVariableRequests.size() > 0 && attractive.size() > 0) {
			if (isProgramFinished()) {
				print("Column selection manager exiting");
				pthread_exit(NULL);
				return NULL; // To prevent a compiler warning
			}
			do {
				var = attractive.at(0);
				attractive.erase(attractive.begin());
			} while (variables[var].status != UNSELECTED && attractive.size() > 0);
			if (variables[var].status != UNSELECTED && attractive.size() == 0)
				break;
			variables[var].status = SELECTED;
			variables[var].basisWhereStatusChanged = kc;
			req = queuedVariableRequests[0];
			queuedVariableRequests.erase(queuedVariableRequests.begin());
			print("x%d gets SELECTED (C5) at basis %d", var, kc);
			MALLOC(varPayload, 1, int);
			*varPayload = var;
			send(req.processName, req.tagToSend, varPayload);
		}
		while (!(message = receiveFromAnyone()).sender) {
			if (isProgramFinished()) {
				print("Column selection manager exiting");
				pthread_exit(NULL);
				return NULL; // To prevent a compiler warning
			}
			sched_yield();
		}
		assert(message.sender[0] == 'I');
		if (!strcmp(message.tag, "I8->C1")) {
			ColSelPivotPayload * payload = (ColSelPivotPayload*)message.payload;
			variables[payload->enteringVar].status = OTHER_SIDE;
			variables[payload->enteringVar].basisWhereStatusChanged = payload->basisNumber;
			variables[payload->leavingVar].status = UNSELECTED;
			variables[payload->leavingVar].basisWhereStatusChanged = payload->basisNumber;
			//print("x%d left basis %d (by %s) and got unselected", payload->leavingVar, payload->basisNumber, message.sender);
			free(payload);
		}
		else if (!strcmp(message.tag, "I9->C2")) {
			if (installAttractiveCandidates(message)) {
				//queuedVariableRequests.Enqueue(new QueuedVariableRequest() { processName = message.sender, tagToSend = "C3->I11" });
				var = attractive.at(0);
				attractive.erase(attractive.begin());
				assert(variables[var].status == UNSELECTED);
				variables[var].status = SELECTED;
				variables[var].basisWhereStatusChanged = kc;
				print("x%d gets SELECTED (C3) at basis %d", var, kc);
				MALLOC(varPayload, 1, int);
				*varPayload = var;
				send(message.sender, "C3->I11", varPayload);
			}
			else {
				req.processName = message.sender;
				req.tagToSend = "C3->I11";
				queuedVariableRequests.push_back(req);
			}
		}
		else if (!strcmp(message.tag, "I4->C4")) {
			QueuedVariableRequest req;
			CandidateIsUnattractivePayload * payload = (CandidateIsUnattractivePayload*)message.payload;
			if (payload->var != -1) {
				variables[payload->var].status = REJECTED;
				variables[payload->var].basisWhereStatusChanged = payload->basisNumber;
				print("x%d gets REJECTED (C5) at basis %d", payload->var, payload->basisNumber);
			}
			free(payload);
			req.processName = message.sender;
			req.tagToSend = "C5->I11";
			queuedVariableRequests.push_back(req);
		}
		else {
			print("Unexpected tag '%s'", message.tag);
			assert(0);
		}
	}
	print("Column selection manager exiting");
	pthread_exit(NULL);
	return NULL; // To prevent a compiler warning
	*/
}
