bool announceProgramFinished() { return true; }
bool isProgramFinished() { return false; }

#include "../PPU/payloadtypes.h"
#include "IterationProcess.h"
//#include "print.h"
#include "common/util.h"
#include "../PPU/common/macros.h"
#include "../PPU/common/myalloc.h"
#include "common/tree.h"
#include "common/linalg.h"
#include "../PPU/common/message.h"
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include "common/tags.h"

using namespace std;

#define EPS_MATRIX 1.0e-8
#define EPSSOL	1.0e-5   /* Zero tolerance for consistent eqns w/dep rows */
#define E_N		200
#define E_NZ	20000
#define SOLVE_USING_LU_EPS	1.0e-14 /* EPS from lueta.c */
#define ITERATION_EPS		1.0e-8	/* EPS from 2phase.c */
#define PICK_NEG_EPS		1.0e-10	/* EPS2 from 2phase.c */
#define GAUSS_ETA_EPS		1.0e-14	/* EPS1 from lueta.c */
#define RATIO_TEST_ETA		1.0e-10	/* EPS1 from 2phase.c */
#define MAX_ITER	1000000

//TODO: Can we find a drand48() implementation for windows?
#ifdef WIN32
#define drand48() (((double)rand())/((double)RAND_MAX))
#endif

#define clock() 0 //TODO

#ifdef QuadPrec
#include "Quad.h"
#define TYPE Quad
#else
#define high(x) (x)
#endif

#define sendAll(a, b) { printf("sendAll not implemented!\n"); fflush(stdout); assert(0); }

static RpcMessage emptyMessage; //TODO
#define receive(x) emptyMessage
#define receiveWithTag(a, b) emptyMessage
#define print(formatString, ...) { printf("I%d: ", this->index); printf(formatString, ## __VA_ARGS__); printf("\n"); fflush(stdout); }

#define streq(a, b) (!strcmp(a, b))

bool IterationProcess::isBasic(int var) {
	return basicflag[var] >= 0;
}

bool IterationProcess::isNonbasic(int var) {
	return basicflag[var] < 0;
}

bool IterationProcess::columnForVariable(int var) {
	if (basicflag[var] >= 0)
		return basicflag[var];
	else
		return -basicflag[var] - 1;
}

void IterationProcess::makeIdentityMatrix(FactoredInverse * inv, int size) {
	int i;
	makeZeroSparseMatrix(&inv->L, size, size);
	makeZeroSparseMatrix(&inv->LT, size, size);
	makeZeroSparseMatrix(&inv->U, size, size);
	makeZeroSparseMatrix(&inv->UT, size, size);
	// All permutation vectors are filled with 0, size-1, ..., 1
	inv->colPerm.resize(size);
	inv->colPerm[0] = 0;
	for (i = 1; i < size; ++i) {
		inv->colPerm[i] = size - i;
	}
	inv->invColPerm = inv->colPerm;
	inv->rowPerm = inv->colPerm;
	inv->invRowPerm = inv->colPerm;
	// Diagonal is only 1s
	inv->diagU.resize(size);
	for (i = 0; i < size; ++i) {
		inv->diagU[i] = 1;
	}
	inv->rank = size;
	inv->basisNumber = 0;
}

void IterationProcess::setup(int index, int numIterationProcesses, SparseMatrix A, TYPE * b, TYPE * c, TYPE f) {
	int i;
	int m = A.rows;
	int N = A.cols;
//	int n = N - m;
	this->index = index;
	this->numIterationProcesses = numIterationProcesses;
	this->comm.setOwner(ITER_PROC(index));
	copySparseMatrix(A, &this->A);
	this->b.resize(m);
	for (i = 0; i < m; ++i)
		this->b[i] = b[i];
	this->c.resize(N);
	for (i = 0; i < N; ++i)
		this->c[i] = c[i];
	this->f = f;
	
	this->currentBasis = 0;
	
	makeIdentityMatrix(&this->factoredInverse, m);
	
	// Reset eta file
	//TODO: Make eta file expandable
	//TODO: Do partial reset only
	this->etaFile.iter = 0; // Updated in refactor
	this->etaFile.size = 0; // Updated at the end of solveUsingLU
	this->etaFile.colPos.push_back(0);
	
	this->tempVector.setRows(m);
	//TODO
	this->internalData_SolveUsingLU.currtagInternal = 1;
	CALLOC( this->internalData_SolveUsingLU.yInternal, m, TYPE);
	CALLOC( this->internalData_SolveUsingLU.tagInternal, m, int);
	
	this->internalData_SolveTransposeUsingLU.currtagInternal = 1;
	CALLOC( this->internalData_SolveTransposeUsingLU.yInternal, m, TYPE);
	CALLOC( this->internalData_SolveTransposeUsingLU.tagInternal, m, int);
	
	this->internalData_Gauss_Eta.currtag = 1;
	CALLOC( this->internalData_Gauss_Eta.a, m, TYPE);
	CALLOC( this->internalData_Gauss_Eta.tag, m, int);
	CALLOC( this->internalData_Gauss_Eta.link, m + 2, int);
	this->internalData_Gauss_Eta.link++; //TODO: Call free() on link-1 !
	
	this->internalData_Gauss_Eta_T.currtag = 1;
	CALLOC( this->internalData_Gauss_Eta_T.a, m, TYPE);
	CALLOC( this->internalData_Gauss_Eta_T.tag, m, int);
	
	this->internalData_Nt_times_y.currtag = 1;
	MALLOC( this->internalData_Nt_times_y.a, N, TYPE);
	CALLOC( this->internalData_Nt_times_y.tag, N, int);
	CALLOC( this->internalData_Nt_times_y.link, N + 2, int);
	this->internalData_Nt_times_y.link++; //TODO: Call free() on link-1 !
	
	this->cumtime = 0;
	this->ocumtime = 0;
}

void IterationProcess::expandEtaFile(int col_out) {
	this->etaFile.newColLocations.push_back(col_out);
	this->etaFile.iter++;
}

//TODO: Make 'peek' functionality so that we can ignore all but the newest inverse
// Repeatedly checks for messages from V with the tag "V2->I1".
// Such a message will contain a more up-to-date LU factorisation of the basis matrix than what this iteration process already has.
// As long as a message is found, the new inverse will be installed (this->factoredInverse will be replaced by the payload of the message)
// and a suitable amount of entries from the front of the eta file will be removed.
void IterationProcess::installNewInverse() {
	RpcMessage message;
	FactoredInverse * inversePayload;
	int i;
	
	while ((message = /*TODO receiveMessage(this->name, "V")*/emptyMessage).sender) {
		assert(message.tag == V2_I1);
		inversePayload = (FactoredInverse*)message.payload;
		//TODO: Restructure loops
		if (inversePayload->basisNumber <= this->factoredInverse.basisNumber) {
			print("Got outdated  B^I for basis %d; currently at basis %d", inversePayload->basisNumber, this->factoredInverse.basisNumber);
		}
		else if (inversePayload->basisNumber > this->currentBasis) {
			print("Got too new B^I for basis %d; process is at basis %d; inverse is at basis %d", inversePayload->basisNumber, this->currentBasis, this->factoredInverse.basisNumber);
		}
		else {
			int numEtaEntriesToRemove;
			int newPos;
			print("Got acceptable B^I for basis %d; process is at basis %d; inverse is at basis %d", inversePayload->basisNumber, this->currentBasis, this->factoredInverse.basisNumber);
			print("installNewInverse is modifying eta file");
			numEtaEntriesToRemove = inversePayload->basisNumber - this->factoredInverse.basisNumber;
			newPos = this->etaFile.colPos[numEtaEntriesToRemove];
			print("%d %d", numEtaEntriesToRemove, newPos);
			this->etaFile.size -= newPos;
			this->etaFile.iter -= numEtaEntriesToRemove;
			// TODO: Can we use some sort of queue structure to make this more efficient?
			for (i = 0; i < this->etaFile.size; ++i) {
				this->etaFile.rowIndices[i] = this->etaFile.rowIndices[i + newPos];
				this->etaFile.values[i] = this->etaFile.values[i + newPos];
			}
			this->etaFile.rowIndices.resize(this->etaFile.size);
			this->etaFile.values.resize(this->etaFile.size);
			for (i = 0; i < this->etaFile.iter + 1; ++i) {
				this->etaFile.colPos[i] = this->etaFile.colPos[i + numEtaEntriesToRemove] - newPos;
			}
			this->etaFile.colPos.resize(this->etaFile.iter + 1);
			assert(this->etaFile.colPos[0] == 0);
			for (i = 0; i < this->etaFile.iter; ++i) {
				this->etaFile.newColLocations[i] = this->etaFile.newColLocations[i + numEtaEntriesToRemove];
			}
			this->etaFile.newColLocations.resize(this->etaFile.iter);
			// Free the space occupied by the previous matrix, copy the message payload and free the original payload
			freeSparseMatrix(&this->factoredInverse.L);
			freeSparseMatrix(&this->factoredInverse.LT);
			freeSparseMatrix(&this->factoredInverse.U);
			freeSparseMatrix(&this->factoredInverse.UT);
			this->factoredInverse = *inversePayload;
			print("basis: %d", this->factoredInverse.basisNumber);
		}
		delete inversePayload;
	}
}

/*************************************************************
 *                            -1  T                           *
 * STEP 2/4: Compute dy  = -(B  N) e                          *
 *                     N            i			              *
 *           where i = col_out                                *
 *************************************************************/
// this->dyN_dualStepDir will be updated.
// The objective function coefficients will change by dyN_dualStepDir * s_dualStepLength (which will be calculated later).
void IterationProcess::FTRAN_calculateDualStepDirection(int col_out) {
	this->tempVector.resize(1);
	this->tempVector.value(0) = -1.0;
	this->tempVector.rowIndex(0) = col_out;
	solveTransposeUsingLU(&this->tempVector);
	Nt_times_y(this->AT, this->basicflag, this->tempVector, this->dyN_dualStepDir);
}

/*************************************************************
 *                          -1                                *
 * STEP 4/2: Compute dx  = B  N e                             *
 *                     B         j                            *
 *           where j = col_in                                 *
 *************************************************************/
// this->dxB_primalStepDir will be updated.
// The right hand side will change by dxB_primalStepDir * t_primalStepLength (which will be calculated later).
// This function calls solveUsingLU(), which as a side effect will partially update the eta file - but etaFile.iter won't be touched, so the change won't be visible until expandEtaFile() is called.
void IterationProcess::FTRAN_calculatePrimalStepDirection(int col_in) {
	int i, j, k;
	j = this->nonbasics[col_in];
	this->dxB_primalStepDir.clear();
	for (i = 0, k = this->A.colPos[j]; k < this->A.colPos[j + 1]; i++, k++) {
		this->dxB_primalStepDir.append(this->A.values[k], this->A.rowIndices[k]);
	}
	solveUsingLU(&this->dxB_primalStepDir);
}

/*************************************************************
 *                                                            *
 * STEP 5: Put       t = x /dx                                *
 *                        i   i                               *
 *                   s = y /dy                                *
 *                        j   j                               *
 *************************************************************/
// Given the leaving and entering columns, computes the factors that dxB_primalStepDir and dyN_dualStepDir must be multiplied by.
void IterationProcess::CHUZR2_calculateStepLengths(int col_in, int col_out, TYPE * t_primalStepLength, TYPE * s_dualStepLength) {
	int k;
	
	/* this is inefficient - it should be fixed */
	for (k = 0; k < this->dxB_primalStepDir.numNonzeroes(); k++)
		if (this->dxB_primalStepDir.rowIndex(k) == col_out)
			break;
	
	*t_primalStepLength = this->x_B[col_out] / this->dxB_primalStepDir.value(k);
	
	/* this is inefficient - it should be fixed */
	for (k = 0; k < this->dyN_dualStepDir.numNonzeroes(); k++)
		if (this->dyN_dualStepDir.rowIndex(k) == col_in)
			break;
	
	*s_dualStepLength = this->y_N[col_in] / this->dyN_dualStepDir.value(k);
}

/*************************************************************
 * STEP 7: Set y  = y  - s dy                                 *
 *              N    N       N                                *
 *                                                            *
 *             y  = s                                         *
 *              i                                             *
 *                                                            *
 *             x  = x  - t dx                                 *
 *              B    B       B                                *
 *                                                            *
 *             x  = t                                         *
 *              j                                             *
 *************************************************************/
void IterationProcess::UPRHS_updatePrimalDualSolutions(int col_in, int col_out, TYPE primalStepLength, TYPE dualStepLength) {
	int i, j, k;
	for (k = 0; k < this->dyN_dualStepDir.numNonzeroes(); k++) {
		j = this->dyN_dualStepDir.rowIndex(k);
		this->y_N[j] -= dualStepLength * this->dyN_dualStepDir.value(k);
	}
	
	this->y_N[col_in] = dualStepLength;
	
	for (k = 0; k < this->dxB_primalStepDir.numNonzeroes(); k++) {
		i = this->dxB_primalStepDir.rowIndex(k);
		this->x_B[i] -= primalStepLength * this->dxB_primalStepDir.value(k);
	}
	
	this->x_B[col_out] = primalStepLength;
}

/*************************************************************
 * STEP 8: Update basis                                       *
 *************************************************************/
void IterationProcess::UPDATE_BASIS(int col_in, int col_out, int callRefactor) {
	int i, j;
	i = this->basics[col_out];
	j = this->nonbasics[col_in];
	this->basics[col_out] = j;
	this->nonbasics[col_in] = i;
	this->basicflag[i] = -col_in - 1;
	this->basicflag[j] = col_out;
	//if (callRefactor)
		//refactor(this->basics, col_out);
}

void IterationProcess::applyIncomingBasisChange(BasisChangePayload * payload) {
	int i;
	print("Applying incoming basis change: to %d", payload->fromBasis + 1);
	assert(payload->fromBasis == this->currentBasis);
	this->dxB_primalStepDir = payload->primalStepDirection;
	this->dyN_dualStepDir = payload->dualStepDirection;
	for (i = 0; i < payload->etaFileSizeDiff; ++i) {
		this->etaFile.values.push_back(payload->newEtaFileValues[i]);
		this->etaFile.rowIndices.push_back(payload->newEtaFileRowIndices[i]);
	}
	this->etaFile.size += payload->etaFileSizeDiff;
	this->etaFile.colPos.push_back(this->etaFile.colPos[this->etaFile.iter] + payload->etaFileSizeDiff);
	this->etaFile.newColLocations.push_back(payload->leavingRow);
	this->etaFile.iter++;
	
	UPDATE_BASIS(payload->enteringCol, payload->leavingRow, 0);
	UPRHS_updatePrimalDualSolutions(payload->enteringCol, payload->leavingRow, payload->primalStepLength, payload->dualStepLength);
	++this->currentBasis;
	delete payload;
}

void IterationProcess::applyIncomingBasisChanges() {
/*TODO
	RpcMessage message;
	while ((message = receiveMessageWithTag(this->name, "I*", "I7->I2I3I10")).sender) {
		assert(!strcmp(message.tag, "I7->I2I3I10"));
		applyIncomingBasisChange(message);
	}
TODO*/
}

// The eta file must be fully updated, including etaFile.iter, before this function is called.
void IterationProcess::broadcastBasisChange(int fromBasis, int col_in, int col_out, int var_in, int var_out, TYPE primalStepLength, TYPE dualStepLength) {
	InvProcPivotPayload invProcPayload;
	ColSelPivotPayload colSelPayload;
	int i, j, k;
	BasisChangePayload * basisChangePayload;
	RpcMessage * messages;
	MALLOC(messages, this->numIterationProcesses - 1, RpcMessage);
	RpcMessage * message = messages;
	for (i = 0; i < this->numIterationProcesses; ++i) {
		if (i == this->index) continue;
		MALLOC(basisChangePayload, 1, BasisChangePayload);
		basisChangePayload->enteringCol = col_in;
		basisChangePayload->leavingRow = col_out;
		basisChangePayload->fromBasis = fromBasis;
		basisChangePayload->primalStepLength = primalStepLength;
		basisChangePayload->dualStepLength = dualStepLength;
		basisChangePayload->primalStepDirection = this->dxB_primalStepDir;
		basisChangePayload->dualStepDirection = this->dyN_dualStepDir;
		basisChangePayload->etaFileSizeDiff = this->etaFile.colPos[this->etaFile.iter] - this->etaFile.colPos[this->etaFile.iter - 1];
		for (j = 0, k = this->etaFile.colPos[this->etaFile.iter - 1]; j < basisChangePayload->etaFileSizeDiff; ++j, ++k) {
			basisChangePayload->newEtaFileValues.push_back(this->etaFile.values[k]);
			basisChangePayload->newEtaFileRowIndices.push_back(this->etaFile.rowIndices[k]);
		}
		print("Sending basis change message to I%d (fromBasis == %d)", i, fromBasis);
		message->sender = ITER_PROC(this->index);
		message->receiver = ITER_PROC(i);
		message->tag = I7_I2I3I10;
		message->payload = basisChangePayload;
		++message;
	}
	//TODO: Need to implement sendAll properly to avoid the race condition described in the report 
	//TODO: Also need to pack the vectors together into one buffer
	sendAll(messages, this->numIterationProcesses - 1);
	delete [] messages;
	invProcPayload.enteringCol = col_in;
	invProcPayload.leavingRow = col_out;
	comm.sendCopy(INV_PROC, I8_V1, &invProcPayload, sizeof(InvProcPivotPayload));
	colSelPayload.enteringVar = var_in;
	colSelPayload.leavingVar = var_out;
	colSelPayload.basisNumber = this->currentBasis;
	comm.sendCopy(COL_SEL_MGR, I8_C1, &colSelPayload, sizeof(ColSelPivotPayload));
}

void IterationProcess::sendAttractiveCandidates(int basisNumber, const SafeVector<AttractiveCandidate> & attractive) {
	int i;
	assert(0);//TODO:
	/*AttractiveCandidatesMessage payload;
	payload->basisNumber = basisNumber;
	for (i = 0; i < attractive.size(); ++i) {
		payload->attractiveVariables.push_back(attractive[i].var);
		//print("%d is attractive", getACV(*attractive, i).col);
	}
	comm.sendCopy(C, I9_C2, payload);*/
}

void IterationProcess::run() {
	SafeSparseVector tempVector;
	
	int col_in; /* entering column; index in 'nonbasics' */
	int col_out; /* leaving column; index in 'basics' */
	int var_out;
	int var_in;

	int iter = 0; /* number of iterations */
	int i, j, k, m, n, N;

	TYPE s_dualStepLength, t_primalStepLength;
	float primal_obj;
	
	TYPE *x; /* primal solution (output) */
	TYPE *y; /* dual solution (output) */
	TYPE *w; /* primal slacks (output) */
	TYPE *z; /* dual slacks (output) */
	
	RpcMessage message;
	InvProcPivotPayload pivotPayload;
	ChuzrOfferPayload chuzrOfferPayload;
	ChuzrAcceptancePayload chuzrAcceptancePayload;
	CandidateIsUnattractivePayload unattractivePayload;

	SafeVector<AttractiveCandidate> attractive;
	
	int keepGoing;
	int finished;
	int accepted;

	seedRandomGeneratorForThisThread();
	
	/*******************************************************************
	 * For convenience, we put...
	 *******************************************************************/
	m = this->A.rows;
	N = this->A.cols;
	n = N - m;
	print("m = %d, n = %d, N = %d, nz = %d", m, n, N, this->A.numNonzeroes);
	
	/*******************************************************************
	 * Read in the Data and initialize the common memory sites.
	 *******************************************************************/
	
	this->x_B.resize(m);
	this->dxB_primalStepDir.setRows(m);
	this->y_N.resize(n);
	this->dyN_dualStepDir.setRows(n);
	tempVector.setRows(m);
	this->nonbasics.resize(n);
	this->basics.resize(m);
	this->basicflag.resize(N);
	this->AT.values = NULL;
	this->AT.rowIndices = NULL;
	this->AT.colPos = NULL;
	nullifySparseMatrix(&this->AT);
	
	/****************************************************************
	 *  Initialization.              				    *
	 ****************************************************************/
	
	transposeSparseMatrix(this->A, &this->AT);
	
	for (j = 0; j < n; j++) {
		this->nonbasics[j] = j;
		this->basicflag[j] = -j - 1;
		this->y_N[j] = MAX(this->c[j],1.0); /* to force dual feasibility */
		this->y_N[j] += drand48(); /* to ensure nondegeneracy   */
	}

	for (i = 0; i < m; i++) {
		this->basics[i] = n + i;
		this->basicflag[n + i] = i;
		this->x_B[i] = this->b[i];
	}
	//TODO: lufac was called here
	dbsolve(this->x_B); /* could be done explicitly in terms of bounds/ranges */
	
	/****************************************************************
	 * 	Begin Phase I (i.e., dual simplex method)
	 ****************************************************************/
	
	CHUZC_findAttractiveCandidates(this->basics, this->x_B, attractive);
	print("index: %d num attractive: %d", this->index, attractive.size());
	if (this->index < attractive.size()) {
		var_out = attractive[this->index].var;
	}
	else {
		var_out = -1;
	}
	
	/****************************************************************
	 * 	Main loop                                                   *
	 ****************************************************************/
	for (iter = 0; iter < MAX_ITER && !isProgramFinished(); iter++) {
		print("iteration %d", iter);
//		assert(iter == this->currentBasis);
		installNewInverse();
		print("here");
		do {
			keepGoing = 0;
			finished = 0;
			
			// Not sure we need to do basis change and FTRAN inside the loop; it might suffice to do them outside
			applyIncomingBasisChanges();
			//TYPE stepLength = 0;
			col_out = -1;
			if (var_out != -1 && isBasic(var_out)) {//TODO: isNonbasic in phase II
				col_out = columnForVariable(var_out);
				print("var_out: %d, col_out: %d", var_out, col_out);
				FTRAN_calculateDualStepDirection(col_out);
			}
			if (var_out == -1 || isNonbasic(var_out) /*TODO: isBasic in phase II*/ || this->x_B[col_out] > 0 /*TODO: y_N in phase II*/) { //TODO: > EPS or <= -EPS ?
				print("x%d is unattractive: col %d, basicflag %d, stepLength %f", var_out, col_out, (var_out == -1 ? -1 : this->basicflag[var_out]), (col_out == -1 ? 0 : this->x_B[col_out]));
				unattractivePayload.var = var_out;
				unattractivePayload.basisNumber = this->currentBasis;
				comm.sendCopy(COL_SEL_MGR, I4_C4, &unattractivePayload, sizeof(CandidateIsUnattractivePayload));
				break; // After the loop, we wait for a new candidate
			}
			else {
				print("x%d is attractive: col %d, basicflag %d, stepLength %f", var_out, col_out, this->basicflag[var_out], this->x_B[col_out]);
			}
			chuzrOfferPayload.basisNumber = this->currentBasis;
			comm.sendCopy(BASIS_MGR, I5_R1, &chuzrOfferPayload, sizeof(ChuzrOfferPayload));
			message.payload = &chuzrAcceptancePayload;
			while (!comm.receiveCopy(BASIS_MGR, ANY_TAG, message, sizeof(ChuzrAcceptancePayload))) {
				if (isProgramFinished()) return;
			}
			assert(message.tag == R2_I6 || message.tag == R3_I6);
			print("R says: %d", chuzrAcceptancePayload.accepted);
			if (chuzrAcceptancePayload.accepted) {
				col_in = CHUZR1_findLeaving_ratioTest(this->dyN_dualStepDir, this->y_N);
				if (col_in == -1) {
					announceProgramFinished();
					print("INFEASIBLE");
					return;/*TODO: return 2; INFEASIBLE */
				}
				var_in = this->nonbasics[col_in]; //TODO: basics in PII
				FTRAN_calculatePrimalStepDirection(col_in);
				CHUZR2_calculateStepLengths(col_in, col_out, &t_primalStepLength, &s_dualStepLength);
				//printRawSparseVector(this->dyN_dualStepDir);
				//printRawSparseVector(this->dxB_primalStepDir);
				expandEtaFile(col_out);
				broadcastBasisChange(this->currentBasis, col_in, col_out, var_in, var_out, t_primalStepLength, s_dualStepLength); //TODO: Be careful - this is different in the two phases!
				//printRawSparseVector(this->dyN_dualStepDir);
				//printRawSparseVector(this->dxB_primalStepDir);
				UPRHS_updatePrimalDualSolutions(col_in, col_out, t_primalStepLength, s_dualStepLength);
				UPDATE_BASIS(col_in, col_out, 1);
				++this->currentBasis;
				
				primal_obj = recalculateObjectiveValues(this->c, this->x_B, this->basics) + this->f;
				print("Iteration %d:\t%14.7e", iter, high(primal_obj));
				
				CHUZC_findAttractiveCandidates(this->basics, this->x_B, attractive);
				if (attractive.size() == 0) {
					finished = 1;
					break; /* ready for Phase II */ //TODO: How to notify the others?
				}
				//col_out = getACV(attractive, 0).col;
				sendAttractiveCandidates(this->currentBasis, attractive);
			}
			else { //Not accepted
				while (!(message = receiveWithTag("I*", "I7->I2I3I10")).sender) {
					if (isProgramFinished()) {
						return;
					}
				}
				assert(message.tag == I7_I2I3I10);
				applyIncomingBasisChange((BasisChangePayload*)message.payload);
				keepGoing = 1;
			}
		} while (keepGoing);
		if (finished)
			break;
		
		while (!(message = receive("C")).sender) {
			if (isProgramFinished()) {
				return;
			}
		}
		assert(message.tag == C3_I11 || message.tag == C5_I11);
		applyIncomingBasisChanges();
		var_out = *(int*)message.payload;
		free(message.payload);
		print("got suggested entering variable: x%d", var_out);
	} /* Phase I loop */
	
	print("End of Phase I");
	if (announceProgramFinished()) {
		return;
	}
	return;
	
	/****************************************************************
	 *   Restore objective function by setting                       *
	 *                -1  T                                          *
	 *         y  = (B  N) c  - c                                    *
	 *          N           B    N                                   *
	 ****************************************************************/
	
	tempVector.clear();
	for (i = 0; i < m; i++) {
		if (ABS(this->c[this->basics[i]]) > ITERATION_EPS) {
			tempVector.append(this->c[this->basics[i]], i);
		}
	}
	
	solveTransposeUsingLU(&tempVector);

	Nt_times_y(this->AT, this->basicflag, tempVector, this->dyN_dualStepDir); // Use dyN_dualStepDir temporarily (this is not the way it is normally used)

	for (j = 0; j < n; j++)
		this->y_N[j] = -this->c[this->nonbasics[j]];
	for (k = 0; k < this->dyN_dualStepDir.numNonzeroes(); k++) {
		j = this->dyN_dualStepDir.rowIndex(k);
		this->y_N[j] += this->dyN_dualStepDir.value(k);
	}

	/****************************************************************
	 *   Begin Phase II (I.e., primal simplex method)                *
	 ****************************************************************/
	
	for (; iter < MAX_ITER; iter++) {
		primal_obj = recalculateObjectiveValues(this->c, this->x_B, this->basics) + this->f;
		print("Iteration %d:\t%14.7e", iter, high(primal_obj));
		
		CHUZC_findAttractiveCandidates(this->nonbasics, this->y_N, attractive);
		if (attractive.size() == 0)
			break; /* optimal */
		var_in = attractive[0].var;
		col_in = columnForVariable(var_in);
		
		installNewInverse();
		
		FTRAN_calculatePrimalStepDirection(col_in);
		
		col_out = CHUZR1_findLeaving_ratioTest(this->dxB_primalStepDir, this->x_B);
		if (col_out == -1) {
			announceProgramFinished();
			print("UNBOUNDED");
			return; /* return 1; UNBOUNDED */
		}
		var_out = this->basics[col_out];

		//TODO: Just for testing messaging w/ InvProc
		pivotPayload.enteringCol = col_in;
		pivotPayload.leavingRow = col_out;
		comm.sendCopy(INV_PROC, I8_V1, &pivotPayload, sizeof(InvProcPivotPayload));
		
		FTRAN_calculateDualStepDirection(col_out);
		
		CHUZR2_calculateStepLengths(col_in, col_out, &t_primalStepLength, &s_dualStepLength);
		
		expandEtaFile(col_out);
		
		UPRHS_updatePrimalDualSolutions(col_in, col_out, t_primalStepLength, s_dualStepLength);
		
		UPDATE_BASIS(col_in, col_out, 1);
		
	} /* End of Phase II */

	primal_obj = recalculateObjectiveValues(this->c, this->x_B, this->basics) + this->f;
	print("%8d   %14.7e     NA    ", iter, high(primal_obj));
	print("End of Phase II \n");
	print("Iterations: %d\nOptimal solution: %14.7e", iter, high(primal_obj));

	/****************************************************************
	 * 	Transcribe solution to x vector and dual solution to y      *
	 ****************************************************************/
	
	MALLOC(x, N, TYPE);
	for (j = 0; j < N; j++)
		x[j] = 0.0;
	for (i = 0; i < m; i++)
		x[this->basics[i]] = this->x_B[i];
	//printTypeArray(x, N);
	
	MALLOC(y, N, TYPE);
	for (j = 0; j < N; j++)
		y[j] = 0.0;
	for (i = 0; i < n; i++)
		y[this->nonbasics[i]] = this->y_N[i];
	
	/****************************************************************
	 * 	Split out slack variables and shift dual variables.
	 ****************************************************************/
	
	MALLOC(z, n, TYPE);
	for (j = 0; j < n; j++)
		z[j] = y[j];
	MALLOC(w, m, TYPE);
	for (i = 0; i < m; i++) {
		y[i] = y[n + i];
		w[i] = x[n + i];
	}
		
	announceProgramFinished();
	print("OPTIMAL");
}

/*-----------------------------------------------------------------+
| Forward/backward solve using LU factorization                    |
| Input:                                                           |
|    m          dimension of array y                               |
|    y          array containing right-hand side                   |
|                                                                  |
|    static global variables (assumed setup by lufac()):           |
|                                                                  |
|    rank       rank of B                                          |
|    L.colPos, L.rowIndices, L, three array sparse representation of L             |
|    UT.colPos,UT.rowIndices,UT.values three array sparse representation of U transpose   |
|               without its diagonal                               |
|    diagU      diagonal entries of U                              |
|    colPerm, invColPerm, rowPerm, invRowPerm                          |
|               column and row permutations and their inverses     |
| Output:                                                          |
|                                          -1                      |
|    y          array containing solution B  y                     |
|                                                                  |
|    integer flag indicating whether system is consistent          |
+-----------------------------------------------------------------*/
// Will update the eta file based on col_out.
int IterationProcess::solveUsingLU(SafeSparseVector * y) {
	int i;
	int k, row, consistent = TRUE;
	TYPE beta;
	TYPE eps;
	
	double starttime, endtime;
	
	Tree tree = {NULL, NULL};
	
	starttime = (double) clock();

	for (k = 0; k < y->numNonzeroes(); k++) {
		i = this->factoredInverse.invRowPerm[y->rowIndex(k)];
		this->internalData_SolveUsingLU.yInternal[i] = y->value(k);
		this->internalData_SolveUsingLU.tagInternal[i] = this->internalData_SolveUsingLU.currtagInternal;
		addtree(&tree, i);
	}

	if (this->factoredInverse.rank < y->numRows())
		eps = EPSSOL * y->maxValue();

	/*------------------------------------------------------+
	 |               -1                                      |
	 |       y  <-  L  y                                    */

	for (i = getfirst(&tree); i < this->factoredInverse.rank && i != -1; i = getnext(&tree)) {
		beta = this->internalData_SolveUsingLU.yInternal[i];
		for (k = this->factoredInverse.L.colPos[i]; k < this->factoredInverse.L.colPos[i + 1]; k++) {
			row = this->factoredInverse.L.rowIndices[k];
			if (this->internalData_SolveUsingLU.tagInternal[row] != this->internalData_SolveUsingLU.currtagInternal) {
				this->internalData_SolveUsingLU.yInternal[row] = 0.0;
				this->internalData_SolveUsingLU.tagInternal[row] = this->internalData_SolveUsingLU.currtagInternal;
				addtree(&tree, row);
			}
			this->internalData_SolveUsingLU.yInternal[row] -= this->factoredInverse.L.values[k] * beta;
		}
	}

	/*------------------------------------------------------+
	 |               -1                                      |
	 |       y  <-  U  y                                    */

	for (i = getlast(&tree); i >= this->factoredInverse.rank && i != -1; i = getprev(&tree)) {
		if (ABS( this->internalData_SolveUsingLU.yInternal[i] ) > eps)
			consistent = FALSE;
		this->internalData_SolveUsingLU.yInternal[i] = 0.0;
	}

	for (; i >= 0; i = getprev(&tree)) {
		beta = this->internalData_SolveUsingLU.yInternal[i] / this->factoredInverse.diagU[i];
		for (k = this->factoredInverse.U.colPos[i]; k < this->factoredInverse.U.colPos[i + 1]; k++) {
			row = this->factoredInverse.U.rowIndices[k];
			if (this->internalData_SolveUsingLU.tagInternal[row] != this->internalData_SolveUsingLU.currtagInternal) {
				this->internalData_SolveUsingLU.yInternal[row] = 0.0;
				this->internalData_SolveUsingLU.tagInternal[row] = this->internalData_SolveUsingLU.currtagInternal;
				addtree(&tree, row);
			}
			this->internalData_SolveUsingLU.yInternal[row] -= this->factoredInverse.U.values[k] * beta;
		}
		this->internalData_SolveUsingLU.yInternal[i] = beta;
	}

	y->clear();
	for (i = getfirst(&tree); i != -1; i = getnext(&tree)) {
		if (ABS(this->internalData_SolveUsingLU.yInternal[i]) > SOLVE_USING_LU_EPS) {
			y->append(this->internalData_SolveUsingLU.yInternal[i], this->factoredInverse.colPerm[i]);
		}
	}

	this->internalData_SolveUsingLU.currtagInternal++;
	killtree(&tree);

	Gauss_Eta(*y);
	
	/*************************************************************
	 * Update etaFile.values and save col_out in etaFile.newColLocations[etaFile.iter]                   *
	 *************************************************************/
	
	print("solveUsingLU is modifying eta file");
	for (i = 0, k = this->etaFile.colPos[this->etaFile.iter]; i < y->numNonzeroes(); i++, k++) {
		this->etaFile.values.push_back(y->value(i));
		this->etaFile.rowIndices.push_back(y->rowIndex(i));
	}

	this->etaFile.size = k;
	this->etaFile.colPos.push_back(k);
	
	endtime = (double) clock();
	this->cumtime += endtime - starttime;

	return consistent;
}

/*-----------------------------------------------------------------+
 | Forward/backward solve using LU factorization                    |
 | Input:                                                           |
 |    y->rows          dimension of array y                               |
 |    y          array containing right-hand side                   |
 |                                                                  |
 |    static global variables (assumed setup by lufac()):           |
 |                                                                  |
 |    rank       rank of B                                          |
 |    L.colPos, L.rowIndices, L, three array sparse representation of L             |
 |    UT.colPos,UT.rowIndices,UT.values three array sparse representation of U transpose   |
 |               without its diagonal                               |
 |    diagU      diagonal entries of U                              |
 |    colPerm, invColPerm, rowPerm, invRowPerm                          |
 |               column and row permutations and their inverses     |
 | Output:                                                          |
 |                                          -T                      |
 |    y          array containing solution B  y                     |
 |                                                                  |
 |    integer flag indicating whether system is consistent         */

int IterationProcess::solveTransposeUsingLU(SafeSparseVector * y) {
	int i;
	int k, row, consistent = TRUE;
	TYPE beta;
	TYPE eps;
	
	double starttime, endtime;
	
	Tree tree = {NULL, NULL};
	
	starttime = (double) clock();

	Gauss_Eta_T(*y);

	for (k = 0; k < y->numNonzeroes(); k++) {
		i = this->factoredInverse.invColPerm[y->rowIndex(k)];
		this->internalData_SolveTransposeUsingLU.yInternal[i] = y->value(k);
		this->internalData_SolveTransposeUsingLU.tagInternal[i] = this->internalData_SolveTransposeUsingLU.currtagInternal;
		addtree(&tree, i);
	}

	if (this->factoredInverse.rank < y->numRows())
		eps = EPSSOL * y->maxValue();

	/*------------------------------------------------------+
	 |               -T                                      |
	 |       y  <-  U  y                                    */

	for (i = getfirst(&tree); i < this->factoredInverse.rank && i != -1; i = getnext(&tree)) {
		beta = this->internalData_SolveTransposeUsingLU.yInternal[i] / this->factoredInverse.diagU[i];
		for (k = this->factoredInverse.UT.colPos[i]; k < this->factoredInverse.UT.colPos[i + 1]; k++) {
			row = this->factoredInverse.UT.rowIndices[k];
			if (this->internalData_SolveTransposeUsingLU.tagInternal[row] != this->internalData_SolveTransposeUsingLU.currtagInternal) {
				this->internalData_SolveTransposeUsingLU.yInternal[row] = 0.0;
				this->internalData_SolveTransposeUsingLU.tagInternal[row] = this->internalData_SolveTransposeUsingLU.currtagInternal;
				addtree(&tree, row);
			}
			this->internalData_SolveTransposeUsingLU.yInternal[row] -= this->factoredInverse.UT.values[k] * beta;
		}
		this->internalData_SolveTransposeUsingLU.yInternal[i] = beta;
	}
	for (i = getlast(&tree); i >= this->factoredInverse.rank && i != -1; i = getprev(&tree)) {
		if (ABS( this->internalData_SolveTransposeUsingLU.yInternal[i] ) > eps)
			consistent = FALSE;
		this->internalData_SolveTransposeUsingLU.yInternal[i] = 0.0;
	}

	/*------------------------------------------------------+
	 |               -T                                      |
	 |       y  <-  L  y                                    */

	for (; i >= 0; i = getprev(&tree)) {
		beta = this->internalData_SolveTransposeUsingLU.yInternal[i];
		for (k = this->factoredInverse.LT.colPos[i]; k < this->factoredInverse.LT.colPos[i + 1]; k++) {
			row = this->factoredInverse.LT.rowIndices[k];
			if (this->internalData_SolveTransposeUsingLU.tagInternal[row] != this->internalData_SolveTransposeUsingLU.currtagInternal) {
				this->internalData_SolveTransposeUsingLU.yInternal[row] = 0.0;
				this->internalData_SolveTransposeUsingLU.tagInternal[row] = this->internalData_SolveTransposeUsingLU.currtagInternal;
				addtree(&tree, row);
			}
			this->internalData_SolveTransposeUsingLU.yInternal[row] -= this->factoredInverse.LT.values[k] * beta;
		}
	}

	y->clear();
	for (i = getfirst(&tree); i != -1; i = getnext(&tree)) {
		if (ABS(this->internalData_SolveTransposeUsingLU.yInternal[i]) > SOLVE_USING_LU_EPS) {
			y->append(this->internalData_SolveTransposeUsingLU.yInternal[i], this->factoredInverse.rowPerm[i]);
		}
	}

	this->internalData_SolveTransposeUsingLU.currtagInternal++;
	killtree(&tree);

	endtime = (double) clock();
	this->cumtime += endtime - starttime;

	return consistent;
}

/*
static void iterationProcess_lu_clo(IterationProcess * proc) {
	freeIV( &this->factoredInverse.rowPerm );
	freeIV( &this->factoredInverse.invRowPerm );
	freeIV( &this->factoredInverse.colPerm );
	freeIV( &this->factoredInverse.invColPerm );
	FREE( this->factoredInverse.L.values );
	FREE( this->factoredInverse.L.rowIndices );
	FREE( this->factoredInverse.L.colPos );
	FREE( this->factoredInverse.UT.values );
	FREE( this->factoredInverse.UT.rowIndices );
	FREE( this->factoredInverse.UT.colPos );
	FREE( this->factoredInverse.diagU );
	freeIV( &this->etaFile.newColLocations );
	freeTV( &this->etaFile.values );
	freeIV( &this->etaFile.rowIndices );
	freeIV( &this->etaFile.colPos );
}
*/

/*-----------------------------------------------------------------+
 | Forward/backward solve using LU factorization                    |
 | Input:                                                           |
 |    m          dimension of array y                               |
 |    y          array containing right-hand side                   |
 |                                                                  |
 |    static global variables (assumed setup by lufac()):           |
 |                                                                  |
 |    rank       rank of B                                          |
 |    L.colPos, L.rowIndices, L.values, three array sparse representation of L.values             |
 |    UT.colPos,UT.rowIndices,UT.values three array sparse representation of U transpose   |
 |               without its diagonal                               |
 |    diagU      diagonal entries of U                              |
 |    colPerm, invColPerm, rowPerm, invRowPerm                          |
 |               column and row permutations and their inverses     |
 | Output:                                                          |
 |                                          -1                      |
 |    y          array containing solution B  y                     |
 |                                                                  |
 |    integer flag indicating whether system is consistent         */

int IterationProcess::dbsolve(SafeVector<TYPE> & y) {
	int i;
	int k, row, consistent = TRUE;
	TYPE beta, *dwork;
	TYPE eps;
	int m = y.size();
	
	double starttime, endtime;

	starttime = (double) clock();

	MALLOC (dwork,m,TYPE);

	if (this->factoredInverse.rank < m)
		eps = EPSSOL * maxv(y);
	for (i = 0; i < m; i++)
		dwork[i] = y[i];
	for (i = 0; i < m; i++)
		y[this->factoredInverse.invRowPerm[i]] = dwork[i];
	/*------------------------------------------------------+
	 |               -1                                      |
	 |       y  <-  L  y                                    */

	for (i = 0; i < this->factoredInverse.rank; i++) {
		beta = y[i];
		for (k = this->factoredInverse.L.colPos[i]; k < this->factoredInverse.L.colPos[i + 1]; k++) {
			row = this->factoredInverse.L.rowIndices[k];
			y[row] -= this->factoredInverse.L.values[k] * beta;
		}
	}

	/*------------------------------------------------------+
	 |               -1                                      |
	 |       y  <-  U  y                                    */

	for (i = m - 1; i >= this->factoredInverse.rank; i--) {
		if (ABS( y[i] ) > eps)
			consistent = FALSE;
		y[i] = 0.0;
	}

	for (i = this->factoredInverse.rank - 1; i >= 0; i--) {
		beta = y[i];
		for (k = this->factoredInverse.UT.colPos[i]; k < this->factoredInverse.UT.colPos[i + 1]; k++) {
			beta -= this->factoredInverse.UT.values[k] * y[this->factoredInverse.UT.rowIndices[k]];
		}
		y[i] = beta / this->factoredInverse.diagU[i];
	}

	for (i = 0; i < m; i++)
		dwork[i] = y[i];
	for (i = 0; i < m; i++)
		y[this->factoredInverse.colPerm[i]] = dwork[i];

	FREE(dwork);

	endtime = (double) clock();
	this->cumtime += endtime - starttime;

	return consistent;
}

void IterationProcess::Gauss_Eta(SafeSparseVector & dx_B) {
	int i, j, k, col, kcol, ii;
	TYPE temp;
	
	if (this->etaFile.iter <= 0) {
		print("Gauss_Eta is returning due to empty eta file");
		return;
	}

	ii = -1;
	for (k = 0; k < dx_B.numNonzeroes(); k++) {
		i = dx_B.rowIndex(k);
		this->internalData_Gauss_Eta.a[i] = dx_B.value(k);
		this->internalData_Gauss_Eta.tag[i] = this->internalData_Gauss_Eta.currtag;
		this->internalData_Gauss_Eta.link[ii] = i;
		ii = i;
	}
	print("Gauss_Eta is reading eta file");
	for (j = 0; j < this->etaFile.iter; j++) {
		assert(this->etaFile.colPos[j] < this->etaFile.colPos[j + 1]);
		col = this->etaFile.newColLocations[j];

		for (k = this->etaFile.colPos[j]; k < this->etaFile.colPos[j + 1]; k++) {
			i = this->etaFile.rowIndices[k];
			if (this->internalData_Gauss_Eta.tag[i] != this->internalData_Gauss_Eta.currtag) {
				this->internalData_Gauss_Eta.a[i] = 0.0;
				this->internalData_Gauss_Eta.tag[i] = this->internalData_Gauss_Eta.currtag;
				this->internalData_Gauss_Eta.link[ii] = i;
				ii = i;
			}
			if (i == col)
				kcol = k;
		}

		temp = this->internalData_Gauss_Eta.a[col] / this->etaFile.values[kcol];
		if (temp != 0.0) {
			for (k = this->etaFile.colPos[j]; k < kcol; k++) {
				i = this->etaFile.rowIndices[k];
				this->internalData_Gauss_Eta.a[i] -= this->etaFile.values[k] * temp;
			}
			this->internalData_Gauss_Eta.a[col] = temp;
			for (k = kcol + 1; k < this->etaFile.colPos[j + 1]; k++) {
				i = this->etaFile.rowIndices[k];
				this->internalData_Gauss_Eta.a[i] -= this->etaFile.values[k] * temp;
			}
		}
	}
	this->internalData_Gauss_Eta.link[ii] = dx_B.numRows();
	this->internalData_Gauss_Eta.currtag++;
	
	dx_B.clear();
	for (i = this->internalData_Gauss_Eta.link[-1]; i < dx_B.numRows(); i = this->internalData_Gauss_Eta.link[i]) {
		if (ABS(this->internalData_Gauss_Eta.a[i]) > GAUSS_ETA_EPS) {
			dx_B.append(this->internalData_Gauss_Eta.a[i], i);
		}
	}
}

/*****************************************************************
 *  Gaussian elimination for Eta transformations which will solve *
 * each system B y = c for y                             	 *
 *****************************************************************/

void IterationProcess::Gauss_Eta_T(SafeSparseVector & vec) {
	int i = -1, j, k, kk = -1, kkk = -1, col;
	TYPE temp;
	
	//print("Gauss_Eta_T is reading eta file");
	for (j = this->etaFile.iter - 1; j >= 0; j--) {
		col = this->etaFile.newColLocations[j];

		for (k = 0; k < vec.numNonzeroes(); k++) {
			i = vec.rowIndex(k);
			if (i == col)
				kk = k;
			this->internalData_Gauss_Eta_T.a[i] = vec.value(k);
			this->internalData_Gauss_Eta_T.tag[i] = this->internalData_Gauss_Eta_T.currtag;
		}

		if (this->internalData_Gauss_Eta_T.tag[col] != this->internalData_Gauss_Eta_T.currtag) {
			kk = vec.numNonzeroes();
			vec.append(0.0, col);
			this->internalData_Gauss_Eta_T.a[col] = 0.0;
			this->internalData_Gauss_Eta_T.tag[col] = this->internalData_Gauss_Eta_T.currtag;
		}
		temp = vec.value(kk);
		for (k = this->etaFile.colPos[j]; k < this->etaFile.colPos[j + 1]; k++) {
			i = this->etaFile.rowIndices[k];
			if (i == col)
				kkk = k;
			if (this->internalData_Gauss_Eta_T.tag[i] == this->internalData_Gauss_Eta_T.currtag) {
				if (i != col) {
					temp -= this->etaFile.values[k] * this->internalData_Gauss_Eta_T.a[i];
				}
			}
		}
		this->internalData_Gauss_Eta_T.currtag++;

		vec.value(kk) = temp / this->etaFile.values[kkk];
	}
}

void IterationProcess::Nt_times_y(SparseMatrix AT, const SafeVector<int> & basicflag, const SafeSparseVector & y, SafeSparseVector & yN) {
	int i, j, jj, k, kk;
	
	jj = -1;
	for (k = 0; k < y.numNonzeroes(); k++) {
		i = y.rowIndex(k);
		for (kk = AT.colPos[i]; kk < AT.colPos[i + 1]; kk++) {
			j = AT.rowIndices[kk];
			if (basicflag[j] < 0) {
				if (this->internalData_Nt_times_y.tag[j] != this->internalData_Nt_times_y.currtag) {
					this->internalData_Nt_times_y.a[j] = 0.0;
					this->internalData_Nt_times_y.tag[j] = this->internalData_Nt_times_y.currtag;
					this->internalData_Nt_times_y.link[jj] = j;
					jj = j;
				}
				this->internalData_Nt_times_y.a[j] += y.value(k) * AT.values[kk];
			}
		}
	}
	this->internalData_Nt_times_y.link[jj] = AT.rows;
	this->internalData_Nt_times_y.currtag++;

	yN.clear();
	for (jj = this->internalData_Nt_times_y.link[-1]; jj < AT.rows; jj = this->internalData_Nt_times_y.link[jj]) {
		if (ABS(this->internalData_Nt_times_y.a[jj]) > EPS_MATRIX) {
			yN.append(this->internalData_Nt_times_y.a[jj], columnForVariable(jj));
		}
	}
}

/*************************************************************
 * STEP 3: Ratio test to find leaving column                  *
 *************************************************************/
int IterationProcess::CHUZR1_findLeaving_ratioTest(const SafeSparseVector & dy, const SafeVector<TYPE> & y) {
	int j, jj = -1, k;
	TYPE min = HUGE_VAL;
	//print("ratio test");
	//printRawSparseVector(dy);
	for (k = 0; k < dy.numNonzeroes(); k++) {
		if (dy.value(k) > RATIO_TEST_ETA) {
			j = dy.rowIndex(k);
			if (y[j] / dy.value(k) < min) {
				min = y[j] / dy.value(k);
				jj = j;
			}
		}
	}
	return jj;
}

TYPE IterationProcess::recalculateObjectiveValues(const SafeVector<TYPE> & c, const SafeVector<TYPE> & x_B, const SafeVector<int> & basics) {
	int i;
	TYPE prod = 0.0;

	for (i = 0; i < x_B.size(); i++) {
		prod += c[basics[i]] * x_B[i];
	}

	return prod;
}

static int compareAttractiveCandidates(const void * a, const void * b) {
	const AttractiveCandidate * candA = (const AttractiveCandidate *)a;
	const AttractiveCandidate * candB = (const AttractiveCandidate *)b;
	if (candA->value < candB->value)
		return -1;
	else if (candA->value > candB->value)
		return 1;
	else
		return 0;
}

/************************************************************************************
 * STEP 1: Pick most negative nonbasic (primal or dual, depending on the parameters) *
 *         'attractive' is assumed to be initialised. 
 ************************************************************************************/
void IterationProcess::CHUZC_findAttractiveCandidates(const SafeVector<int> & columnToVariableMapping, const SafeVector<TYPE> & x, SafeVector<AttractiveCandidate> & attractive) {
	int i;
	TYPE best = 0;
	attractive.clear();
	for (i = 0; i < x.size(); i++) {
		if (x[i] < best) {
			best = x[i];
		}
		if (x[i] < -PICK_NEG_EPS) {
			AttractiveCandidate cand;
			cand.var = columnToVariableMapping[i];
			cand.value = x[i];
			attractive.push_back(cand);
		}
	}
	printf("best candidate: %e\n", best);fflush(stdout);
	if (attractive.size() > 0) {
		qsort(&attractive[0], attractive.size(), sizeof(AttractiveCandidate), compareAttractiveCandidates);
	}
}
