#ifndef TYPES_H
#define TYPES_H

#include "sparse.h"
#include "SafeVector.h"

#define VECTOR_WIDTH 4
#define ALIGN_CACHE_WIDTH 128
#define ALIGN_CACHE_LOG2 7
#define ALIGN_BUS_LOG2 7
#define ALIGN_BUS_WIDTH 128
#define ALIGN_QUAD_WIDTH 16
#define ALIGN_QUAD_LOG2 4
#define ALIGNED_CACHE __attribute__ ((aligned(ALIGN_CACHE_WIDTH)))
#define ALIGNED_QUAD __attribute__ ((aligned(ALIGN_QUAD_WIDTH)))
#define ROUND_UP_MULTIPLE(x, m) (((x) + (m) - 1) / (m) * (m))
#define PADDING(actualSize, alignmentWidth) char padding[ROUND_UP_MULTIPLE(actualSize, alignmentWidth) - (actualSize)]

// All of these are IEEE 754 quiet NaNs, which means that they can't accidentally be interpreted as (or collide with) valid floating point numbers.
#define SIMPLEX_MBOX_OPTIMAL 0xffffffff
#define SIMPLEX_MBOX_UNBOUNDED 0xfffffffe
#define SIMPLEX_MBOX_LEAVING_IS_HERE 0xfffffffd
#define SIMPLEX_MBOX_LEAVING_IS_ELSEWHERE 0xfffffffc

//#define INFINITY (__builtin_inff())
#define INFINITY 3.4E38f

typedef unsigned int uint;

struct ParameterContext {
	int rows;
	int cols;
	int physicalCols;
	int startingRow;
	int numContainedRows;
	int spuIndex;
	float * dataOrigin;
	float * ppuPivotRow;
	PADDING(6 * sizeof(int) + 2 * sizeof(float*), ALIGN_QUAD_WIDTH);
};

union Value32 {
	uint uintValue;
	int intValue;
	float floatValue;
};

enum SimplexResult {
	SUBOPTIMAL,
	OPTIMAL,
	UNBOUNDED,
	CYCLING, 
	FAILURE
};


typedef struct {
	SparseMatrix L;
	SparseMatrix LT;
	SparseMatrix U;
	SparseMatrix UT;
	
	SafeVector<TYPE> diagU;
	SafeVector<int> colPerm;
	SafeVector<int> invColPerm;
	SafeVector<int> rowPerm;
	SafeVector<int> invRowPerm;
	int rank;
	
	int basisNumber;
} FactoredInverse;

#endif

