#define SPE_CODE

#include "BasisChangeManager.h"
#include "ColumnSelectionManager.h"
#include "IterationProcess.h"
#include <spu_intrinsics.h>
#include <spu_mfcio.h>
#include <libmisc.h>
#include <stdio.h>
#include "../common/types.h"
#include "rpc.h"
#include "common/sparse.h"

volatile ParameterContext context ALIGNED_QUAD;

ppe_funcptr sendRpcMessage ALIGNED16;
ppe_funcptr sendRpcMessageBatch ALIGNED16;
ppe_funcptr receiveRpcMessage ALIGNED16;
ppe_funcptr getDimensions ALIGNED16;
ppe_funcptr getIntArrays ALIGNED16;
ppe_funcptr getTypeArrays ALIGNED16;

//int main(unsigned long long spuId __attribute__ ((unused)), unsigned long long parameter) {
extern "C" int main (int argc, char * argv[]) { // Using extern "C" because of CML
	int rank, size;
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	printf("MPI: %d/%d\n", rank, size);fflush(stdout);
	
	sendRpcMessage = cellmsg_accept_rpc();
	sendRpcMessageBatch = cellmsg_accept_rpc();
	receiveRpcMessage = cellmsg_accept_rpc();
	getDimensions = cellmsg_accept_rpc();
	getIntArrays = cellmsg_accept_rpc();
	getTypeArrays = cellmsg_accept_rpc();
	
	int dimensions[4] ALIGNED16;
	cellmsg_rpc(getDimensions, 	// Function to call 
				NULL, 0, CML_BYTE_SWAP_NOT_NEEDED, 		// No input
				dimensions, 4 * sizeof(int), CML_BYTE_SWAP_NOT_NEEDED); // Output
	int m = dimensions[0];
	int n = dimensions[1];
	int numNonzeroes = dimensions[2];
	int N = m + n;
	
	//TODO: Merge BasisChange and ColSel
	if (rank == 0) {
		BasisChangeManager basisChangeManager;
		basisChangeManager.run();
	}
	else if (rank == 1) {
		ColumnSelectionManager columnSelectionManager;
		columnSelectionManager.setup(m, n);
		columnSelectionManager.run();
	}
	else {
		if (rank != 2) {
			MPI_Finalize();
			return 0;
		}
		/**/
		int intCount = numNonzeroes + (N + 1);		// A.rowIndices, A.colPos
		int typeCount = 1 + m + N + numNonzeroes;	// f, b, c, A.values
		int intBytes = ROUND_UP_MULTIPLE(intCount * sizeof(int), ALIGN_BUS_WIDTH);
		int typeBytes = ROUND_UP_MULTIPLE(typeCount * sizeof(TYPE), ALIGN_BUS_WIDTH);
		int * intArrays = (int*)malloc_align(intBytes, ALIGN_BUS_LOG2);
		TYPE * typeArrays = (TYPE*)malloc_align(typeBytes, ALIGN_BUS_LOG2);
		printf("Getting %d ints and %d types (numNonzeroes: %d)\n", intCount, typeCount, numNonzeroes);fflush(stdout);
		cellmsg_rpc(getIntArrays, 
					NULL, 0, CML_BYTE_SWAP_NOT_NEEDED, 
					intArrays, intBytes, CML_BYTE_SWAP_NOT_NEEDED);
		cellmsg_rpc(getTypeArrays, 
					NULL, 0, CML_BYTE_SWAP_NOT_NEEDED, 
					typeArrays, typeBytes, CML_BYTE_SWAP_NOT_NEEDED);
		/*int i = 0;
		while (true) {
			i += 1024;
			if (!malloc(1024)) {
				MPI_Finalize();
				return 0;
			}
			printf("%d\n", i);
		}*/
		/*
		if (rank == 2) {
			for (int i = 0; i < intCount; ++i) {
				printf("%d ", intArrays[i]);
			}
			printf("\n");fflush(stdout);
			for (int i = 0; i < typeCount; ++i) {
				printf("%e ", typeArrays[i]);
			}
			printf("\n");fflush(stdout);
		}
		*/
		TYPE f = typeArrays[0];
		TYPE * b = typeArrays + 1;
		TYPE * c = typeArrays + 1 + m;
		SparseMatrix A;
		A.rows = m;
		A.cols = N;
		A.numNonzeroes = numNonzeroes;
		A.rowIndices = intArrays;
		A.colPos = intArrays + numNonzeroes;
		A.values = typeArrays + 1 + m + N;
		
		IterationProcess iterationProcess;
		iterationProcess.setup(rank - 2, size - 2, A, b, c, f);
		free_align(intArrays);
		free_align(typeArrays);
		iterationProcess.run();
	}
	MPI_Finalize();
	return 0;
}
