#include "rpc.h"
#include <libmisc.h>
#include "common/types.h"
#include <assert.h>
#include <string.h>

void Communication::sendCopy(int receiver, int tag, void * payload, int payloadSize) {
	int bufferSize = payloadSize + 4 * sizeof(int);
	int roundedBufferSize = ROUND_UP_MULTIPLE(bufferSize, ALIGN_BUS_WIDTH);
	int * buffer = (int*)malloc_align(roundedBufferSize, ALIGN_BUS_LOG2);
	buffer[SENDER_OFFSET] = this->owner;
	buffer[RECEIVER_OFFSET] = receiver;
	buffer[TAG_OFFSET] = tag;
	buffer[BUFFER_SIZE_OFFSET] = bufferSize;
	memcpy(buffer + PAYLOAD_OFFSET, payload, payloadSize);
	printf("sendCopy: %d %d %d, %d %d %d %p\n", buffer[0], buffer[1], buffer[2], payloadSize, bufferSize, roundedBufferSize, buffer);
	cellmsg_rpc(sendRpcMessage, 
				buffer, roundedBufferSize, CML_BYTE_SWAP_NOT_NEEDED, 
				NULL, 0, CML_BYTE_SWAP_NOT_NEEDED);
	free_align(buffer);
}

bool Communication::receiveCopy(int sender, int tag, RpcMessage & message, int payloadSize) {
	int bufferSize = payloadSize + 4 * sizeof(int);
	int roundedBufferSize = ROUND_UP_MULTIPLE(bufferSize, ALIGN_BUS_WIDTH);
	int * buffer = (int*)malloc_align(roundedBufferSize, ALIGN_BUS_LOG2);
	buffer[SENDER_OFFSET] = sender;
	buffer[RECEIVER_OFFSET] = this->owner;
	buffer[TAG_OFFSET] = tag;
	cellmsg_rpc(receiveRpcMessage, 
				buffer, 4 * sizeof(int), CML_BYTE_SWAP_NOT_NEEDED, 
				buffer, roundedBufferSize, CML_BYTE_SWAP_NOT_NEEDED);
	bool success = buffer[SENDER_OFFSET] >= 0; // receiveRpcMessage will put -1 at element 0 if there is no matching message
	if (success) {
		message.sender = buffer[SENDER_OFFSET];
		message.receiver = buffer[RECEIVER_OFFSET];
		message.tag = buffer[TAG_OFFSET];
		assert(buffer[BUFFER_SIZE_OFFSET] >= payloadSize);
		memcpy(message.payload, buffer + PAYLOAD_OFFSET, payloadSize);
	}
	free_align(buffer);
	return success; 
}

void Communication::sendBuffer(int receiver, int tag, void * buffer, int totalSize) {
	assert(0);
}
