#ifndef RPC_H
#define RPC_H

#include <spu_intrinsics.h> // Need to include this one before mpi.h, because otherwise extern "C" will screw up the files that mpi.h includes (might need to include even more of those files here if we start using them)
extern "C" {
	#include <mpi.h>
}
#include "common/message.h"

#define ALIGNED16  __attribute__((aligned(16)))
#define ALIGN_BUS_LOG2 7
#define ALIGN_BUS_WIDTH 128

// Indices into the int buffer used for transmitting messages
#define SENDER_OFFSET 0
#define RECEIVER_OFFSET 1
#define TAG_OFFSET 2
#define BUFFER_SIZE_OFFSET 3
#define PAYLOAD_OFFSET 4

// Give files that include this one access to the function pointers defined in main.cpp
extern ppe_funcptr sendRpcMessage ALIGNED16;
extern ppe_funcptr sendRpcMessageBatch ALIGNED16;
extern ppe_funcptr receiveRpcMessage ALIGNED16;

class Communication {
private:
	int owner;
public:
	void setOwner(int owner) { this->owner = owner; }
	void sendCopy(int receiver, int tag, void * data, int dataSize);
	void sendBuffer(int receiver, int tag, void * buffer, int totalSize);
	bool receiveCopy(int sender, int tag, RpcMessage & message, int payloadSize);
};

#endif
