#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "../simpo/sparse.h"

void VECTOR_INIT(VECTOR_NAME * vector) {
	vector->size = 0;
	vector->totalSize = 16;
	vector->array = (VECTOR_TYPE*)malloc(vector->totalSize * sizeof(VECTOR_TYPE));
}

void VECTOR_RESIZE(VECTOR_NAME * vector, int size) {
	vector->size = size;
	vector->totalSize = size;
	vector->array = (VECTOR_TYPE*)realloc(vector->array, vector->totalSize * sizeof(VECTOR_TYPE));
}

void VECTOR_FREE(VECTOR_NAME * vector) {
	vector->size = 0;
	vector->totalSize = 0;
	free(vector->array);
	vector->array = NULL;
}

void VECTOR_APPEND(VECTOR_NAME * vector, VECTOR_TYPE element) {
	if (vector->size == vector->totalSize) {
		vector->totalSize *= 2;
		vector->array = (VECTOR_TYPE*)realloc(vector->array, vector->totalSize * sizeof(VECTOR_TYPE));
	}
	vector->array[vector->size++] = element;
}

VECTOR_TYPE VECTOR_GET(VECTOR_NAME vector, int index) {
	assert(index >= 0 && index < vector.size);
	return vector.array[index];
}

void VECTOR_SET(VECTOR_NAME vector, int index, VECTOR_TYPE value) {
	assert(index >= 0 && index < vector.size);
	vector.array[index] = value;
}

void VECTOR_REMOVE_AT(VECTOR_NAME * vector, int index) {
	int i;
	assert(index >= 0 && index < vector->size);
	--vector->size;
	for (i = index; i < vector->size; ++i) {
		vector->array[i] = vector->array[i + 1];
	}
}

// destination is assumed not to point to anything that must be freed
void VECTOR_COPY(VECTOR_NAME source, VECTOR_NAME * destination) {
	destination->size = source.size;
	destination->totalSize = source.size;
	allocateAndCopyArray(source.array, (void**)&destination->array, source.size, sizeof(VECTOR_TYPE));
}
