#include <stdlib.h>

#ifdef QuadPrec
#include "Quad.h"
#define TYPE Quad
#else
#define high(x) (x)
#endif

#include "linalg.h"
#include "myalloc.h"
#include "macros.h"
#include "../simpo/sparse.h"

using namespace std;

/*---------------------------------------------------------------+
|  inner product between n-vectors x and y                      */

TYPE dotprod( TYPE *x, TYPE *y, int n)
{
        int i;
        TYPE dotprod=0.0e0;

        for (i=0; i<n; i++) dotprod += x[i]*y[i];

        return (dotprod);
}

/*---------------------------------------------------------------+
|  y = basis submatrix of (a,ka,ia) times x                     */

void bmx( int m, TYPE *a, int *ka, int *ia, int *basis,
          TYPE *x, TYPE *y)
{
        int i,j,k;

        for (i=0; i<m; i++) y[i] = 0.0e0;
        for (i=0; i<m; i++) {
                j = basis[i];
                for (k=ka[j]; k<ka[j+1]; k++)
                        y[ia[k]] += a[k]*x[i];
        }
}

/*---------------------------------------------------------------+
|  y = basis submatrix of (a,ka,ia) transpose times x           */

void btmx( int m, TYPE *a, int *ka, int *ia, int *basis,
          TYPE *x, TYPE *y)
{
        int i,j,k;

        for (i=0; i<m; i++) y[i] = 0.0e0;
        for (i=0; i<m; i++) {
                j = basis[i];
                for (k=ka[j]; k<ka[j+1]; k++)
                        y[i] += a[k]*x[ia[k]];
        }
}

/*---------------------------------------------------------------+
|  y = sparse matrix (a,ka,ia) times x                          */

void smx( int m, int n, TYPE *a, int *ka, int *ia, TYPE *x, TYPE *y)
{
        int i,j,k;

        for (i=0; i<m; i++) y[i] = 0.0e0;
        for (j=0; j<n; j++)
                for (k=ka[j]; k<ka[j+1]; k++)
                        y[ia[k]] += a[k]*x[j];
}

// Takes care of setting all values in *AT_ptr, including (re)allocation. The pointers in *AT_ptr must either point to valid memory or be NULL.
void transposeSparseMatrix(SparseMatrix A, SparseMatrix * AT_ptr) {
	AT_ptr->cols = A.rows;
	AT_ptr->rows = A.cols;
	AT_ptr->numNonzeroes = A.numNonzeroes;
	MALLOC_OR_REALLOC(AT_ptr->values, AT_ptr->numNonzeroes, TYPE);
	MALLOC_OR_REALLOC(AT_ptr->rowIndices, AT_ptr->numNonzeroes, int);
	MALLOC_OR_REALLOC(AT_ptr->colPos, AT_ptr->cols + 1, int);
	transposeMatrix(A.rows, A.cols, A.colPos, A.rowIndices, A.values, AT_ptr->colPos, AT_ptr->rowIndices, AT_ptr->values);
}

/*---------------------------------------------------------------+
|  (kat,iat,at) = transpose of (ka,ia,a)                        */

void transposeMatrix( int m, int n, int *ka, int *ia, TYPE *a,
        int *kat,int *iat, TYPE *at
        )
{
        int i,j,k,row,addr;
        int *iwork;

        CALLOC( iwork, m, int );

        for (k=0; k<ka[n]; k++) {
                row = ia[k];
                iwork[row]++;
        }
        	
        kat[0] = 0;
        for (i=0; i<m; i++) {
                kat[i+1] = kat[i] + iwork[i];
                iwork[i] = 0;
        }
        	
        for (j=0; j<n; j++) {
                for (k=ka[j]; k<ka[j+1]; k++) {
                        row = ia[k];
                        addr = kat[row] +iwork[row];
                        iwork[row]++;
                        iat[addr] = j;
                        at[addr]  = a[k];
                }
        }
        	
        FREE( iwork );
}

/*---------------------------------------------------------------+
|  compute componentwise maximum of n-vector x                  */

TYPE maxv( TYPE *x, int n)
{
        int i;
        TYPE maxv=0.0e0;

        for (i=0; i<n; i++) maxv = MAX(maxv, ABS(x[i]));

        return (maxv);
}

TYPE maxv(const vector<TYPE> & x)
{
        int i;
        TYPE maxv=0.0e0;

        for (i=0; i<x.size(); i++) maxv = MAX(maxv, ABS(x.at(i)));

        return (maxv);
}
