#include "communication.h"
#include "invertprocessor.h"
#include "print.h"
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>

#define NAME "R"

static int currentBasis = 0;

static void send(const char * receiver, const char * tag, void * payload) {
	Message msg = {NAME, receiver, tag, payload};
	sendMessage(msg);
}

static Message receiveFromAnyone() {
	return receiveMessageFromAnyone(NAME);
}

extern int enablePrint; // Defined in iterationprocess.c

#define print(formatString, ...) if (enablePrint) { printFromProcess(NAME, formatString , ## __VA_ARGS__); }

void * runBasisChangeManager(void * parameters) {
	Message message;
	const char * result;
	int * payload;
	int incomingBasis;
	print("Basis change manager starting");
	while (!isProgramFinished()) {
		while (!(message = receiveFromAnyone()).sender) {
			if (isProgramFinished()) {
				print("Basis change manager exiting");
				pthread_exit(NULL);
				return NULL; // To prevent a compiler warning
			}
			sched_yield();
		}
		assert(!strcmp(message.tag, "I5->R1"));
		incomingBasis = *(int*)message.payload;
		free(message.payload);
		payload = (int*)malloc(sizeof(int));
		if (incomingBasis == currentBasis) {
			*payload = 1;
			send(message.sender, "R2->I6", payload);
			++currentBasis;
			result = "accepted";
		}
		else {
			*payload = 0;
			send(message.sender, "R3->I6", payload);
			result = "refused";
		}
		print("%s offers chuzr on basis %d... %s", message.sender, incomingBasis, result);
	}
	print("Basis change manager exiting");
	pthread_exit(NULL);
	return NULL; // To prevent a compiler warning
}
