#include "communication.h"
#include "print.h"
#include "../common/genericvectors.h"
#include "payloadtypes.h"
#include "invertprocessor.h"
#include <assert.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
using namespace std;

#define NAME "C"

static char * VARIABLE_STATUS_NAMES[] = {"OTHER_SIDE", "UNSELECTED", "SELECTED", "REJECTED", "UNDEFINED"};

static VariableInfo * variables;
static vector<int> attractive;
static QueuedVariableRequestVector queuedVariableRequests;
static int m;
static int n;
static int kc;

static void send(const char * receiver, const char * tag, void * payload) {
	Message msg = {NAME, receiver, tag, payload};
	sendMessage(msg);
}

static Message receiveFromAnyone() {
	return receiveMessageFromAnyone(NAME);
}

extern int enablePrint; // Defined in iterationprocess.c

#define print(formatString, ...) if (enablePrint) { printFromProcess(NAME, formatString , ## __VA_ARGS__); }

void setupColumnSelectionManager(int _m, int _n) {
	int i;
	m = _m;
	n = _n;
	kc = 0;
	initQVRV(&queuedVariableRequests);
	variables = (VariableInfo*)malloc(sizeof(VariableInfo) * (m + n));
	//TODO: This must be switched between the phases
	for (i = 0; i < n; ++i) { // nonbasic
		variables[i].status = OTHER_SIDE;
		variables[i].basisWhereStatusChanged = -1;
	}
	for (i = 0; i < m; ++i) { // basic
		variables[n + i].status = UNSELECTED;
		variables[n + i].basisWhereStatusChanged = -1;
	}
}

bool installAttractiveCandidates(Message message) {
	int i;
	AttractiveCandidatesMessage msg = *(AttractiveCandidatesMessage*)message.payload;
	delete (AttractiveCandidatesMessage*)message.payload;
	assert(msg.attractiveVariables.size() >= 1);
	print("received %d candidates from %s, which is at basis %d; currently at basis %d", msg.attractiveVariables.size(), message.sender, msg.basisNumber, kc);
	if (msg.basisNumber > kc) {
		int installCount = 0;
		attractive.clear();
		for (i = 0; i < msg.attractiveVariables.size(); ++i) {
			int cand = msg.attractiveVariables.at(i);
			if (!((variables[cand].status == SELECTED || variables[cand].status == REJECTED) && variables[cand].basisWhereStatusChanged >= msg.basisNumber)) {
				++installCount;
				attractive.push_back(cand);
				variables[cand].status = UNSELECTED;
				variables[cand].basisWhereStatusChanged = msg.basisNumber;
			}
			else {
				print("rejected x%d because it got %s at basis %d", cand, VARIABLE_STATUS_NAMES[variables[cand].status], variables[cand].basisWhereStatusChanged);
			}
		}
		print("Installed %d candidates", installCount);
		kc = msg.basisNumber;
		return true;
	}
	else {
		print("Rejected the candidates");
		return false;
	}
}

void * runColumnSelectionManager(void * parameters) {
	Message message;
	int var;
	int * varPayload;
	QueuedVariableRequest req;
	print("Column selection manager starting");
	
	while (!isProgramFinished()) {
		while (queuedVariableRequests.size > 0 && attractive.size() > 0) {
			if (isProgramFinished()) {
				print("Column selection manager exiting");
				pthread_exit(NULL);
				return NULL; // To prevent a compiler warning
			}
			do {
				var = attractive.at(0);
				attractive.erase(attractive.begin());
			} while (variables[var].status != UNSELECTED && attractive.size() > 0);
			if (variables[var].status != UNSELECTED && attractive.size() == 0)
				break;
			variables[var].status = SELECTED;
			variables[var].basisWhereStatusChanged = kc;
			req = getQVRV(queuedVariableRequests, 0);
			removeAtQVRV(&queuedVariableRequests, 0);
			print("x%d gets SELECTED (C5) at basis %d", var, kc);
			varPayload = (int*)malloc(sizeof(int));
			*varPayload = var;
			send(req.processName, req.tagToSend, varPayload);
		}
		while (!(message = receiveFromAnyone()).sender) {
			if (isProgramFinished()) {
				print("Column selection manager exiting");
				pthread_exit(NULL);
				return NULL; // To prevent a compiler warning
			}
			sched_yield();
		}
		assert(message.sender[0] == 'I');
		if (!strcmp(message.tag, "I8->C1")) {
			ColSelPivotPayload * payload = (ColSelPivotPayload*)message.payload;
			variables[payload->enteringVar].status = OTHER_SIDE;
			variables[payload->enteringVar].basisWhereStatusChanged = payload->basisNumber;
			variables[payload->leavingVar].status = UNSELECTED;
			variables[payload->leavingVar].basisWhereStatusChanged = payload->basisNumber;
			//print("x%d left basis %d (by %s) and got unselected", payload->leavingVar, payload->basisNumber, message.sender);
			free(payload);
		}
		else if (!strcmp(message.tag, "I9->C2")) {
			if (installAttractiveCandidates(message)) {
				//queuedVariableRequests.Enqueue(new QueuedVariableRequest() { processName = message.sender, tagToSend = "C3->I11" });
				var = attractive.at(0);
				attractive.erase(attractive.begin());
				assert(variables[var].status == UNSELECTED);
				variables[var].status = SELECTED;
				variables[var].basisWhereStatusChanged = kc;
				print("x%d gets SELECTED (C3) at basis %d", var, kc);
				varPayload = (int*)malloc(sizeof(int));
				*varPayload = var;
				send(message.sender, "C3->I11", varPayload);
			}
			else {
				req.processName = message.sender;
				req.tagToSend = "C3->I11";
				appendQVRV(&queuedVariableRequests, req);
			}
		}
		else if (!strcmp(message.tag, "I4->C4")) {
			QueuedVariableRequest req;
			CandidateIsUnattractivePayload * payload = (CandidateIsUnattractivePayload*)message.payload;
			if (payload->var != -1) {
				variables[payload->var].status = REJECTED;
				variables[payload->var].basisWhereStatusChanged = payload->basisNumber;
				print("x%d gets REJECTED (C5) at basis %d", payload->var, payload->basisNumber);
			}
			free(payload);
			req.processName = message.sender;
			req.tagToSend = "C5->I11";
			appendQVRV(&queuedVariableRequests, req);
		}
		else {
			print("Unexpected tag '%s'", message.tag);
			assert(0);
		}
	}
	print("Column selection manager exiting");
	pthread_exit(NULL);
	return NULL; // To prevent a compiler warning
}
