#include "../common/message.h"
#include "../common/genericvectors.h"
#include "communication.h"
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include "print.h"

Message createMessage(const char * sender, const char * receiver, const char * tag, void * payload) {
	Message message;
	message.sender = sender;
	message.receiver = receiver;
	message.tag = tag;
	message.payload = payload;
	return message;
}

static MessageVector messages;
static pthread_mutex_t queueLock;
static int printSends = 1;
static int printReceives = 1;

void initCommunication() {
	pthread_mutex_init(&queueLock, NULL);
	initMV(&messages);
}

void sendMessage(Message message) {
	pthread_mutex_lock(&queueLock);
	appendMV(&messages, message);
	pthread_mutex_unlock(&queueLock);
	if (printSends) {
		printFromProcess("COMM", "%s sends to %s (tag: '%s', payload %p) (queue: %d)", message.sender, message.receiver, message.tag, message.payload, messages.size);
	}
}

void sendAll(Message * messagesToBeSent, int numMessages) {
	int i;
	pthread_mutex_lock(&queueLock);
	for (i = 0; i < numMessages; ++i) {
		appendMV(&messages, messagesToBeSent[i]);
		if (printSends) {
			printFromProcess("COMM", "%s sends to %s (tag: '%s', payload %p) (queue: %d)", messagesToBeSent[i].sender, messagesToBeSent[i].receiver, messagesToBeSent[i].tag, messagesToBeSent[i].payload, messages.size);
		}
	}
	pthread_mutex_unlock(&queueLock);
}

// Returns a message with sender == null if no message is available
Message receiveMessageFromAnyone(const char * receiver) {
	return receiveMessageWithTag(receiver, NULL, NULL);
}

// Returns a message with sender == null if no message is available
Message receiveMessage(const char * receiver, const char * sender) {
	return receiveMessageWithTag(receiver, sender, NULL);
}

Message receiveMessageWithTag(const char * receiver, const char * sender, const char * tag) {
	Message msg;
	int i;
	pthread_mutex_lock(&queueLock);
	for (i = 0; i < messages.size; ++i) {
		msg = getMV(messages, i);
		if (!strcmp(msg.receiver, receiver) && (!sender || !strcmp(msg.sender, sender) || (!strcmp(sender, "I*") && msg.sender[0] == 'I')) && (!tag || !strcmp(msg.tag, tag))) {
			removeAtMV(&messages, i);
			if (printReceives) {
				printFromProcess("COMM", "%s receives from %s (tag: '%s', payload: %p) (queue: %d)", msg.receiver, msg.sender, msg.tag, msg.payload, messages.size);
			}
			pthread_mutex_unlock(&queueLock);
			return msg;
		}
	}
	pthread_mutex_unlock(&queueLock);
	msg.sender = NULL;
	return msg;
}

void printMessages() {
	int i;
	pthread_mutex_lock(&queueLock);
	printFromProcess("COMM", "Remaining messages:");
	for (i = 0; i < messages.size; ++i) {
		printFromProcess("COMM", "%s %s %s", getMV(messages, i).sender, getMV(messages, i).receiver, getMV(messages, i).tag);
	}
	fflush(stdout);
	pthread_mutex_unlock(&queueLock);
}
