#ifndef ITERATIONPROCESS_H
#define ITERATIONPROCESS_H

#include <pthread.h>
#include "sparse.h"
#include "../common/genericvectors.h"

typedef struct {
	SparseMatrix L;
	SparseMatrix LT;
	SparseMatrix U;
	SparseMatrix UT;
	
	SafeVector<TYPE> diagU;
	SafeVector<int> colPerm;
	SafeVector<int> invColPerm;
	SafeVector<int> rowPerm;
	SafeVector<int> invRowPerm;
	int rank;
	
	int basisNumber;
} FactoredInverse;

/* The eta file seems to be stored as one sparse matrix, 
 * where each column represents one eta matrix. */
typedef struct {
	int iter;				/* Number of eta matrices */
	int size;				/* Number of entries in 'values' */
	SafeVector<TYPE> values;			/* Nonzero values */
	SafeVector<int> rowIndices;		/* Row indices - seem to be based on 'colPerm' */
	SafeVector<int> colPos;			/* Indices into 'values' and 'rowIndices' telling where each eta matrix column begins */
	SafeVector<int> newColLocations;	/* Column location for each eta matrix (leaving column for corresponding pivot) */
} EtaFile;

typedef struct {
	TYPE * yInternal;
	int * tagInternal;
	int currtagInternal;
} InternalData_SolveUsingLU;

typedef struct {
	TYPE * yInternal;
	int * tagInternal;
	int currtagInternal;
} InternalData_SolveTransposeUsingLU;

typedef struct {
	TYPE * a;
	int * tag;
	int * link;
	int currtag;
} InternalData_Gauss_Eta;

typedef struct {
	TYPE * a;
	int * tag;
	int currtag;
} InternalData_Gauss_Eta_T;

typedef struct {
	TYPE * a;
	int * tag;
	int * link;
	int currtag;
} InternalData_Nt_times_y;

typedef struct {
	int index;
	int numIterationProcesses;
	const char * name;
	char ** iterationProcessNames;
	pthread_t thread; // It's safe to copy pthread_t objects: http://newsgroups.derkeiler.com/Archive/Comp/comp.programming.threads/2007-07/msg00283.html
	
	SparseMatrix A;
	SparseMatrix AT;
	SafeVector<TYPE> b;	/* right-hand side */
	SafeVector<TYPE> c;	/* objective coefficients */
	TYPE f;		/* objective function shift */
	SafeVector<TYPE> x_B;	/* primal basics */
	SafeVector<TYPE> y_N;	/* dual nonbasics */
	
	int currentBasis;
	
	FactoredInverse factoredInverse;
	
	int     e_iter; /* number of iterations since last refactorization */
	
	EtaFile etaFile;
	
	SafeSparseVector dyN_dualStepDir;
	SafeSparseVector dxB_primalStepDir;
	SafeSparseVector tempVector;
	
	SafeVector<int> basics;		/* list of basic variable indices */
	SafeVector<int> nonbasics;	/* list of non-basic variable indices */
	SafeVector<int> basicflag;	/* negative if nonbasic, nonnegative if basic */
	
	InternalData_SolveUsingLU internalData_SolveUsingLU;
	InternalData_SolveTransposeUsingLU internalData_SolveTransposeUsingLU;
	InternalData_Gauss_Eta internalData_Gauss_Eta;
	InternalData_Gauss_Eta_T internalData_Gauss_Eta_T;
	InternalData_Nt_times_y internalData_Nt_times_y;
	
	double cumtime; //TODO: This one was used both by the functions that were moved to invertprocessor.c and to interationprocess.c
	double ocumtime;
} IterationProcess;

typedef struct {
	int var;
	TYPE value;
} AttractiveCandidate;

void setupIterationProcess(IterationProcess * process, int index, int numIterationProcesses, SparseMatrix A, TYPE * b, TYPE * c, TYPE f);
void * runIterationProcess(void * parameters);

#endif
