#include "sparse.h"
#include "../common/linalg.h"
#include <stdlib.h>
#include <stdio.h>

void nullifySparseMatrix(SparseMatrix * matrix) {
	matrix->rows = 0;
	matrix->cols = 0;
	matrix->numNonzeroes = 0;
	matrix->rowIndices = NULL;
	matrix->colPos = NULL;
	matrix->values = NULL;
}

void makeZeroSparseMatrix(SparseMatrix * matrix, int rows, int cols) {
	matrix->rows = rows;
	matrix->cols = cols;
	matrix->numNonzeroes = 0;
	matrix->rowIndices = (int*)malloc(0);
	matrix->values = (TYPE*)malloc(0);
	matrix->colPos = (int*)malloc((cols + 1) * sizeof(int));
	memset(matrix->colPos, 0, (cols + 1) * sizeof(int));
}

void nullifySparseVector(SparseVector * vector) {
	vector->rows = 0;
	vector->numNonzeroes = 0;
	vector->rowIndices = NULL;
	vector->values = NULL;
}

// does not free destination
void allocateAndCopyArray(const void * source, void ** destination, int numElements, size_t elementSize) {
	*destination = malloc(numElements * elementSize);
	memcpy(*destination, source, numElements * elementSize);
}

void copySparseMatrix(SparseMatrix source, SparseMatrix * destination) {
	destination->rows = source.rows;
	destination->cols = source.cols;
	destination->numNonzeroes = source.numNonzeroes;
	allocateAndCopyArray(source.rowIndices, (void**)&destination->rowIndices, source.numNonzeroes, sizeof(int)); //TODO: Are we guaranteed this size?
	allocateAndCopyArray(source.colPos, (void**)&destination->colPos, (source.cols + 1), sizeof(int));
	allocateAndCopyArray(source.values, (void**)&destination->values, source.numNonzeroes, sizeof(TYPE));
}

void copySparseVector(SparseVector source, SparseVector * destination) {
	destination->rows = source.rows;
	destination->numNonzeroes = source.numNonzeroes;
	//TODO: Allocate 'rows', but only copy 'numNonzeroes'
	allocateAndCopyArray(source.rowIndices, (void**)&destination->rowIndices, source.rows, sizeof(int)); //TODO: Are we guaranteed this size?
	allocateAndCopyArray(source.values, (void**)&destination->values, source.rows, sizeof(TYPE));
}

void printRawSparseMatrix(SparseMatrix matrix) {
	int i;
	fflush(stdout);
	printf("%d x %d (%d):\n", matrix.rows, matrix.cols, matrix.numNonzeroes);
	for (i = 0; i < matrix.cols + 1; ++i)
		printf("%d ", matrix.colPos[i]);
	printf("\n");
	for (i = 0; i < matrix.numNonzeroes; ++i)
		printf("%d:%e ", matrix.rowIndices[i], matrix.values[i]);
	printf("\n");
	fflush(stdout);
}

void printRawSparseVector(SparseVector vector) {
	int i;
	fflush(stdout);
	printf("%d (%d):", vector.rows, vector.numNonzeroes);
	for (i = 0; i < vector.numNonzeroes; ++i)
		printf(" %d:%e", vector.rowIndices[i], vector.values[i]);
	printf("\n");
	fflush(stdout);
}

void printTypeArray(TYPE * array, int size) {
	int i;
	fflush(stdout);
	printf("%d:", size);
	for (i = 0; i < size; ++i) {
		printf(" %e", array[i]);
	}
	printf("\n");
	fflush(stdout);
}

void printIntArray(int * array, int size) {
	int i;
	fflush(stdout);
	printf("%d:", size);
	for (i = 0; i < size; ++i) {
		printf(" %d", array[i]);
	}
	printf("\n");
	fflush(stdout);
}

/*
void initSparseMatrix(SparseMatrix * matrix, int rows, int cols) {
	matrix->rows = rows;
	matrix->cols = cols;
	matrix->numNonzeroes = 0;
	matrix->rowIndices = NULL;
	matrix->colPos = NULL;
	matrix->values = NULL;
}
*/

void initSparseVector(SparseVector * vector, int rows) {
	vector->numNonzeroes = 0;
	vector->rowIndices = (int*)malloc(sizeof(int) * rows);
	vector->rows = rows;
	vector->values = (TYPE*)malloc(sizeof(TYPE) * rows);
}


void freeSparseMatrix(SparseMatrix * matrix) {
	free(matrix->colPos);
	free(matrix->rowIndices);
	free(matrix->values);
	nullifySparseMatrix(matrix);
}

void freeSparseVector(SparseVector * vector) {
	free(vector->rowIndices);
	free(vector->values);
	nullifySparseVector(vector);
}

TYPE SafeSparseVector::maxValue() const {
	return maxv(values);
}
