#ifndef SPARSE_H
#define SPARSE_H

#include <string.h>
#include <vector>

typedef struct {
	int rows;
	int cols;
	int numNonzeroes;
	int * rowIndices;
	int * colPos;
	TYPE * values;
} SparseMatrix;

typedef struct {
	int rows;
	int numNonzeroes;
	int * rowIndices;
	TYPE * values;
} SparseVector;

class SafeSparseVector {
private:
	int rows;
	std::vector<int> rowIndices;
	std::vector<TYPE> values;
public:
	int numNonzeroes() const {
		return values.size();
	}
	int numRows() const {
		return rows;
	}
	void resize(int size) {
		rowIndices.resize(size);
		values.resize(size);
	}
	void setRows(int rows) {
		this->rows = rows;
	}
	void clear() {
		rowIndices.clear();
		values.clear();
	}
	void append(TYPE value, int rowIndex) {
		values.push_back(value);
		rowIndices.push_back(rowIndex);
	}
	TYPE & value(int i) {
		return values.at(i);
	}
	TYPE value(int i) const {
			return values.at(i);
	}
	int & rowIndex(int i) {
		return rowIndices.at(i);
	}
	int rowIndex(int i) const {
		return rowIndices.at(i);
	}
	TYPE maxValue() const;
};

void nullifySparseMatrix(SparseMatrix * matrix);
void makeZeroSparseMatrix(SparseMatrix * matrix, int rows, int cols);
void allocateAndCopyArray(const void * source, void ** destination, int numElements, size_t elementSize);
void copySparseMatrix(SparseMatrix source, SparseMatrix * destination);
void copySparseVector(SparseVector source, SparseVector * destination);
void nullifySparseVector(SparseVector * vector);
void printRawSparseMatrix(SparseMatrix matrix);
void printRawSparseVector(SparseVector vector);
void printTypeArray(TYPE * array, int size);
void printIntArray(int * array, int size);

//void initSparseMatrix(SparseMatrix * matrix, int rows, int cols);
void initSparseVector(SparseVector * vector, int rows);
void freeSparseMatrix(SparseMatrix * matrix);
void freeSparseVector(SparseVector * vector);

template<typename T>
class SafeVector : public std::vector<T> {
public:
	T & operator [] (int index) {
		return this->at(index);
	}
	T operator [] (int index) const {
		return this->at(index);
	}
};

#endif
