#!/usr/bin/python
from sys import stdin, stderr

class Equation:
	comparator = ""
	constant = 0
	values = {}
	name = ""

	def __init__(self, comparator, constant, name):
		self.comparator = comparator
		self.constant = constant
		self.values = {}
		self.name = name



class Bound:
	variable = ""
	lower = 0
	upper = None
	free = False
	fixed = False

	def __init__(self, variable):
		self.variable = variable

variableCodeNames = {}

def truncate(name):
	global variableCodeNames
 	if len(name) <= 8:
 		return name
 	else:
		if variableCodeNames.has_key(name):
			return variableCodeNames[name]
		else:
			codeName = "v" + str(len(variableCodeNames))
			variableCodeNames[name] = codeName
			return codeName

def printCodeNames():
	global variableCodeNames
	if len(variableCodeNames) > 0:
		stderr.write("Some variable names have been changed:\n")
		stderr.write("New\tOriginal\n")
	for name in variableCodeNames:
		stderr.write(variableCodeNames[name] + "\t" + name + "\n")

def expand(string, length):
	if len(string) > length:
		raise ValueError("string too long")
	return string + " " * (length - len(string))

class LP:
	pos = 0
	lines = []
	variables = {}
	equations = []
	variableList = []
	bounds = []
	direction = "max"

	def __init__(self):
		lines = []
		variables = {}
		equations = []
		variableList = []
	
	def printMPS(self):
		print "NAME          UNKNOWN"
		print "ROWS"
		for eq in self.equations:
			if eq.comparator == "=":
				print " E ",
			elif eq.comparator[0] == "<":
				print " L ",
			elif eq.comparator[0] == ">":
				print " G ",
			elif eq.comparator == "obj":
				print " N ",
			else:
				raise NameError("Illegal comparator: " + eq.comparator)
			print expand(truncate(eq.name), 8)
		print "COLUMNS"
		for var in self.variableList:
			for eq in self.equations:
				if eq.values.has_key(var):
					line = expand("    " + truncate(var), 14) + truncate(eq.name)
					print expand(line, 24) + str(eq.values[var])
		print "RHS"
		for eq in self.equations:
			if eq.constant != 0:
				print expand("    B         " + truncate(eq.name), 24) + str(eq.constant)
		print "BOUNDS"
		for bound in self.bounds:
			if bound.free:
				print " FR BOUND     " + truncate(bound.variable)
			elif bound.fixed:
				print expand(" FX BOUND     " + truncate(bound.variable), 24) + str(bound.upper)
			else:
				if bound.lower != 0:
					print expand(" LO BOUND     " + truncate(bound.variable), 24) + str(bound.lower)
				if bound.upper != None:
					print expand(" UP BOUND     " + truncate(bound.variable), 24) + str(bound.upper)
		print "ENDATA"
	
	def parseObjective(self):
		tokens = self.lines[self.pos]
		self.pos += 1
		self.parseEquation(tokens, 1)
	
	def parseEquation(self, tokens, isObjective):
		if tokens[1] != '+' and tokens[1] != '-':
			tokens.insert(1, '+')
		if isObjective:
			eq = Equation("obj", 0, "OBJ")
		else:
			eq = Equation(tokens[-2], float(tokens[-1]), tokens[0][:-1])
		self.equations.append(eq)
		i = 1
		limit = len(tokens) - 1 if isObjective else len(tokens) - 3
		while i < limit:
			if tokens[i] == '-':
				sign = -1
			elif tokens[i] == '+':
				sign = 1
			else:
				print "Illegal sign on line", self.pos, ":", tokens
			if isObjective and self.direction == "max":
				sign *= -1
			try:
				value = float(tokens[i + 1])
				i += 2
			except ValueError:
				value = 1
				i += 1
			name = tokens[i]
			self.addVariable(name)
			eq.values[name] = sign * value
			i += 1

	def parseEquations(self):
		while 1:
			tokens = self.lines[self.pos]
			if tokens[0][-1] != ':': break
			self.pos += 1
			self.parseEquation(tokens, 0)

	def addVariable(self, name):
		if not self.variables.has_key(name):
			self.variables[name] = len(self.variables)
			self.variableList.append(name)

	def parseBounds(self):
		while 1:
			tokens = self.lines[self.pos]
			if len(tokens) == 1: break
			self.pos += 1
			if len(tokens) == 2 and tokens[1] == "Free":
				bound = Bound(tokens[0])
				bound.free = True
				self.bounds.append(bound)
			elif len(tokens) == 3:
				bound = Bound(tokens[0])
				if tokens[1][0] == "<":
					bound.upper = float(tokens[2])
				elif tokens[1][0] == ">":
					bound.lower = float(tokens[2])
				elif tokens[1][0] == "=":
					bound.fixed = True
					bound.upper = float(tokens[2])
				else:
					raise NameError("Illegal bound type")
				self.bounds.append(bound)
			elif len(tokens) == 5:
				bound = Bound(tokens[2])
				bound.lower = float(tokens[0])
				bound.upper = float(tokens[4])
				self.bounds.append(bound)
			else:
				print "Unrecognised bounds line:", self.pos, ":", tokens
	
	def parse(self):
		for line in stdin:
			tokens = line.split()
			if len(tokens) == 0 or tokens[0] == '\\': continue
			self.lines.append(tokens)
		self.pos = 0
		while self.pos < len(self.lines):
			if self.lines[self.pos][0] == "Maximize":
				self.direction = "max"
				self.pos += 1
				self.parseObjective()
			elif self.lines[self.pos][0] == "Minimize":
				self.direction = "min"
				self.pos += 1
				self.parseObjective()
			elif self.lines[self.pos][0] == "Subject":
				self.pos += 1
				self.parseEquations()
			elif self.lines[self.pos][0] == "Bounds":
				self.pos += 1
				self.parseBounds()
			else:
				self.pos += 1

lp = LP()
lp.parse()
lp.printMPS()
printCodeNames()
