#ifdef USE_GMP
	#include <gmpxx.h>
#endif
#include "gmpInterop.h"
using namespace std;

#ifdef USE_GMP
// Only handles floats using regular notation (no 0x, no E) and integers.
// Accepts numbers < 1 without a leading zero.
// The string cannot be empty or contain any spaces.
static void parseRational(const string & str, mpq_class & number) {
	//bool negative = (str[0] == '-');
	size_t dotIndex = str.find('.');
	if (dotIndex == string::npos) {
		number = str;
		number.canonicalize();
		return;
	}
	string noDot = str.substr(0, dotIndex) + str.substr(dotIndex + 1);
//	size_t firstNonzeroIndex = noDot.find_first_not_of("-0");
//	if (firstNonzeroIndex == string::npos) {
//		number = 0;
//		return;
//	}
//	string noLeadingZeroes = negative ? "-" : "";
//	noLeadingZeroes += noDot.substr(firstNonzeroIndex);
	mpq_class numerator(noDot, 10);
	int power = str.size() - dotIndex - 1;
	mpq_class denominator = 1;
	while (power--) {
		denominator *= 10;
	}
	number = numerator / denominator;
	number.canonicalize();
}
#endif

void readNumber(istream & in, TYPE & number) {
#ifdef USE_GMP
	string str;
	in >> str;
	parseRational(str, number);
#else
	in >> number;
#endif
}

void printNumber(const TYPE & number) {
#ifdef USE_GMP
	cout << number.get_d();
#else
	cout << number;
#endif
}

void printNumberFull(const TYPE & number) {
#ifdef USE_GMP
	cout << number.get_d() << " (approx); " << number << " (exact)";
#else
	cout << number;
#endif
}

/*
int main() {
	mpq_class a;
	parseRational("0.00001", a);
	cout << a << endl;
	parseRational("10.00001", a);
	cout << a << endl;
	parseRational("12.00000", a);
	cout << a << endl;
	parseRational(".02001", a);
	cout << a << endl;
	parseRational("-.5", a);
	cout << a << endl;
	parseRational("-0.00001", a);
	cout << a << endl;
	parseRational("-120", a);
	cout << a << endl;
	parseRational("-0.0", a);
	cout << a << endl;
	parseRational("-0.0", a);
	cout << a << endl;
	parseRational("0", a);
	cout << a << endl;
	parseRational("-0", a);
	cout << a << endl;
	parseRational("32.1234", a);
	cout << a << endl;
	return 0;
}
*/

