#ifndef PARTIALMATRIX_H
#define PARTIALMATRIX_H

class PartialMatrix {
public:
	PartialMatrix(int, int, int, int, int);
	PartialMatrix(int rows, int cols, int physicalCols, int startRow, int containedRows, float * data);
	PartialMatrix(int rows, int cols, int physicalCols, int startRow, int containedRows, bool identity);
	PartialMatrix(const PartialMatrix &);
	~PartialMatrix();
	inline int getRows() const { return rows; }
	inline int getCols() const { return cols; }
	inline int getPhysicalCols() const { return physicalCols; }
	inline int getStartRow() const { return startRow; }
	inline int getContainedRows() const { return containedRows; }
	inline int getRowLimit() const { return startRow + containedRows; }
	inline volatile float * getData() { return data; }
	inline volatile float & operator () (int r, int c) { return data[(r - startRow) * physicalCols + c]; }
	inline float operator () (int r, int c) const { return data[(r - startRow) * physicalCols + c]; }
	void multiplyRow(int row, float factor);
	void addRows(float factor, int sourceRow, int destinationRow);
	void addRows(float factor, float * sourceRow, int destinationRow);
	void swapRows(int firstRow, int secondRow);
	void print();
private:
	void init(int rows, int cols, int physicalCols, int startRow, int containedRows);
	
	int rows;
	int cols;
	int physicalCols;
	int startRow;
	int containedRows;
	volatile float * data;
};

#endif
