#define SPE_CODE

#include <spu_intrinsics.h>
#include <spu_mfcio.h>
#include <libmisc.h>
#include <stdio.h>
# include "../types.h"
#include "PartialMatrix.h"
#include "SpuTableauSimplex.h"

#define MAX_DMA_SIZE 16384

volatile ParameterContext context ALIGNED_QUAD;

void safeDMA(volatile float * localStoreAddress, unsigned int effectiveAddress, unsigned int size, unsigned int tagId, unsigned int command) {
	int remainingSize = size;
	while (remainingSize > 0) {
		spu_mfcdma32(localStoreAddress, effectiveAddress, (remainingSize > MAX_DMA_SIZE ? MAX_DMA_SIZE : remainingSize), tagId, command);
		remainingSize -= MAX_DMA_SIZE;
		effectiveAddress += MAX_DMA_SIZE;
		localStoreAddress += MAX_DMA_SIZE / sizeof(float);
		(void)spu_mfcstat(MFC_TAG_UPDATE_ALL);
	}
}

int main(unsigned long long spuId __attribute__ ((unused)), unsigned long long parameter) {
	uint tagId;
	if ((tagId = mfc_tag_reserve()) == MFC_TAG_INVALID) {
		printf("ERROR (%llx): unable to reserve a tag\n", spuId);
		return 1;
	}
	spu_writech(MFC_WrTagMask, -1);

	// Fetch the context and wait
	spu_mfcdma32((void *)&context, (uint)parameter, sizeof(ParameterContext), tagId, MFC_GET_CMD);
	(void)spu_mfcstat(MFC_TAG_UPDATE_ALL);
	
	int submatrixLength = context.numContainedRows * context.physicalCols;
	PartialMatrix matrix(context.rows, context.cols, context.physicalCols, context.startingRow, context.numContainedRows);
	safeDMA(matrix.getData(), (uint)(context.dataOrigin + context.startingRow * context.physicalCols), submatrixLength * sizeof(float), tagId, MFC_GETB_CMD); // MFC_GETB_CMD can be used for a barrier get, which will wait until previous puts have completed - http://publib.boulder.ibm.com/infocenter/systems/scope/syssw/index.jsp?topic=/eiccj/tutorial/cbet_3optimz.html

	while (SpuTableauSimplex::iterate(matrix, context.ppuPivotRow, tagId, context.spuIndex) == SUBOPTIMAL);
	
	safeDMA(matrix.getData(), (uint)(context.dataOrigin + context.startingRow * context.physicalCols), submatrixLength * sizeof(float), tagId, MFC_PUT_CMD);
	
	return 0;
}

