#ifndef TYPES_H
#define TYPES_H

#define VECTOR_WIDTH 4
#define ALIGN_CACHE_WIDTH 128
#define ALIGN_CACHE_LOG2 7
#define ALIGN_QUAD_WIDTH 16
#define ALIGN_QUAD_LOG2 4
#define ALIGNED_CACHE __attribute__ ((aligned(ALIGN_CACHE_WIDTH)))
#define ALIGNED_QUAD __attribute__ ((aligned(ALIGN_QUAD_WIDTH)))
#define ROUND_UP_MULTIPLE(x, m) (((x) + (m) - 1) / (m) * (m))
#define PADDING(actualSize, alignmentWidth) char padding[ROUND_UP_MULTIPLE(actualSize, alignmentWidth) - (actualSize)] // Adds a char array to the end of a struct in order to make the struct take up a certain number of bytes - the first parameter is the combined size of the other struct memebers, and the second parameter is the byte boundary on which the struct size should be divisible.

// Sentinel values used when an SPE wants to notify the PPE that it did not find a good leaving or entering variable, and when the PPE wants to let the SPEs know who should perform the pivot. All of these are IEEE 754 quiet NaNs, which means that they can't accidentally be interpreted as (or collide with) valid floating point numbers.
#define SIMPLEX_MBOX_OPTIMAL 0xffffffff
#define SIMPLEX_MBOX_UNBOUNDED 0xfffffffe
#define SIMPLEX_MBOX_LEAVING_IS_HERE 0xfffffffd
#define SIMPLEX_MBOX_LEAVING_IS_ELSEWHERE 0xfffffffc

//#define INFINITY (__builtin_inff())
#ifndef INFINITY
	#define INFINITY 3.4E38f
#endif

typedef unsigned int uint;

struct ParameterContext {
	int rows;
	int cols;
	int physicalCols;
	int startingRow;
	int numContainedRows;
	int spuIndex;
	float * dataOrigin;
	float * ppuPivotRow;
	PADDING(6 * sizeof(int) + 2 * sizeof(float*), ALIGN_QUAD_WIDTH);
};

// For interpreting a 32 bit pattern in different ways.
union Value32 {
	uint uintValue;
	int intValue;
	float floatValue;
};

enum SimplexResult {
	SUBOPTIMAL,
	OPTIMAL,
	UNBOUNDED,
	CYCLING, 
	FAILURE
};

#endif

