#include "TableauSimplex.h"
#include "gmpInterop.h"
#include <cmath>
#include <vector>

using namespace std;

void TableauSimplex::pivot(Matrix & tableau, std::vector<int> & basic, std::vector<int> & nonbasic, int leaving, int entering) {
	cout << "Pivoting: x" << basic[leaving - 1] << " (row " << leaving << ") leaves, x" << nonbasic[entering - 1] << " (column " << entering << ") enters" << endl;
	TYPE xFactor = tableau(leaving, entering);
	int leavingLabel = basic[leaving - 1];
	basic[leaving - 1] = nonbasic[entering - 1];
	nonbasic[entering - 1] = leavingLabel;

	// Cancel out occurrences of the entering variable
	for (int i = 0; i < tableau.getRows(); ++i) {
		if (i == leaving) continue;
		TYPE factor = -tableau(i, entering) / xFactor;
		TYPE savedColVal = tableau(i, entering);
		tableau.addRows(leaving, i, factor);
		tableau(i, entering) = savedColVal / xFactor;
	}

	tableau.multiplyRow(leaving, -1 / xFactor);
	tableau(leaving, entering) = 1 / xFactor;
}



SimplexResult TableauSimplex::iteration(Matrix & tableau, vector<int> & basic, vector<int> & nonbasic) {
	int n = tableau.getCols() - 1, m = tableau.getRows() - 1;

	// Find entering variable by searching the objective function (row 0) for a positive coefficient (disregard the constant in column 0)
	int entering = -1;
	for (int j = 1; j <= n; ++j) {
		if (tableau(0, j) > 0 && (entering == -1 || tableau(0, j) > tableau(0, entering) || (tableau(0, j) == tableau(0, entering) && nonbasic[j - 1] < nonbasic[entering - 1])))
			entering = j;
	}
	if (entering == -1)
		return OPTIMAL;

	// Find leaving variable by searching the column of the entering variable and determine the strictest bound
	int leaving = -1;
	TYPE smallestRatio = -1; // Keep the compiler from complaining about uninitialised variables
	for (int i = 1; i <= m; ++i) {
		if (tableau(i, entering) >= 0)
			continue;
		TYPE ratio = -tableau(i, 0) / tableau(i, entering); // The "right hand side", tableau(i, 0), is always nonnegative, and we only get here if tableau(i, entering) is negative, so 'ratio' will be nonnegative
		if (leaving == -1 || ratio < smallestRatio || (ratio == smallestRatio && basic[i - 1] < basic[leaving - 1])) {
			smallestRatio = ratio;
			leaving = i;
		}
	}
	if (leaving == -1)
		return INFEASIBLE_OR_UNBOUNDED;

	pivot(tableau, basic, nonbasic, leaving, entering);

	return SUBOPTIMAL;
}

